import pyvista as pv
from midas_civil import Node,Element,Group
import numpy as np

class Visual:
    n_snap = 0  # Goes from 1 to snaps taken
    cur_snap = 0
    visual_info = {}
    plotter = pv.Plotter(window_size=[700,400],title="Sumit's Visualiser TEST")
    
    # plotter.camera.clipping_range = (0.000001, 1e6)
    plotter.set_background("white")
    plotter.enable_parallel_projection()
    plotter.show_axes()
    # plotter.ren_win.SetBorders(0)
    plotter.ren_win.SetPosition(600,100)

    first_launch = True

    min_z = 0
    min_x = 0
    max_x = 0
    min_y = 0
    max_y = 0

    
    toggle_NodeIDs = True
    toggle_Beam = True
    toggle_Plate = True
    toggle_GroupNodeIDs = False
    toggle_Grid = False

def take_snapshot():
    Visual.n_snap+=1
    Visual.cur_snap+=1

    points = []
    point_ids = []

    for nd in Node.nodes:
        points.append([nd.X,nd.Y,nd.Z])
        point_ids.append(nd.ID)
    
    point_Group_ids = ["." for _ in point_ids]

    for grup in Group.Structure.Groups:
        if grup.NLIST !=[]:
            idx = 0
            for nID in grup.NLIST:
                n_idx = point_ids.index(nID)
                point_Group_ids[n_idx] = str(idx)
                idx+=1



    lines_point_pair =[]
    lines_id_map = []
    Qplates_point_pair = []
    Qplates_id_map =[]
    Tplates_point_pair = []
    Tplates_id_map = []


    for elm in Element.elements:
        if elm.TYPE in ['BEAM','TRUSS']:
            n1 = point_ids.index(elm.NODE[0])
            n2 = point_ids.index(elm.NODE[1])
            lines_point_pair.append([2,n1,n2])
            lines_id_map.append(elm.ID)

        elif elm.TYPE == 'PLATE':
            n1 = point_ids.index(elm.NODE[0])
            n2 = point_ids.index(elm.NODE[1])
            n3 = point_ids.index(elm.NODE[2])
            if elm._NPOINT == 3 :
                Tplates_point_pair.append([3,n1,n2,n3])
                Tplates_id_map.append(elm.ID)
            elif elm._NPOINT == 4 :
                # print("4 noded element")
                n4 = point_ids.index(elm.NODE[3])
                Qplates_point_pair.append([4,n1,n2,n3,n4])
                Qplates_id_map.append(elm.ID)

    Visual.visual_info[str(Visual.n_snap)] = {
        "POINTS" : points,
        "POINT_IDS" : point_ids,
        "POINT_GROUP_IDS" : point_Group_ids,
        "LINE_POINTS" : lines_point_pair,
        "LINE_IDS" : lines_id_map,
        "TRI_POINTS" : Tplates_point_pair,
        "TRI_IDS" : Tplates_id_map,
        "QUAD_POINTS" : Qplates_point_pair,
        "QUAD_IDS" : Qplates_id_map
    }

    # print("TAKING SNAP SHOT ................")
    # print(Visual.n_snap)
    # print(len(points))
    # print(len(Qplates_id_map))
    # print("---------- D O N E -------------")

def changeDataBack(checked):
        # print("DATA Change....")
        Visual.cur_snap = max(Visual.cur_snap-1,1)
        # print(Visual.cur_snap)
        displayWindow()

def changeDataForw(checked):
        # print("DATA Change....")
        Visual.cur_snap = min(Visual.cur_snap+1,Visual.n_snap)
        # print(Visual.cur_snap)
        displayWindow()

def displayWindow():
    if Visual.first_launch == True:
        take_snapshot()
        dPlotter = showVisual(Visual.plotter)
        Visual.first_launch = False
        
        dPlotter.view_xy()
        dPlotter.show()
    else:
        Visual.plotter.clear_actors()
        dPlotter = showVisual(Visual.plotter)
        dPlotter.render()


def showVisual(plotter):

    if Visual.first_launch:
        min_z = 0
        min_x = 0
        max_x = 0
        min_y = 0
        max_y = 0
        for nd in Node.nodes:
            min_z = min(min_z,nd.Z)
            min_x = min(min_x,nd.X)
            max_x = max(max_x,nd.X)
            min_y = min(min_y,nd.Y)
            max_y = max(max_y,nd.Y)
        min_z = int(min_z)
        min_x = int(min_x)
        max_x = int(max_x)
        min_y = int(min_y)
        max_y = int(max_y)

        Visual.min_x = min_x
        Visual.max_x = max_x
        Visual.min_y = min_y
        Visual.max_y = max_y
        Visual.min_z = min_z
    else :
        min_x = Visual.min_x
        max_x = Visual.max_x
        min_y = Visual.min_y
        max_y = Visual.max_y
        min_z = Visual.min_z


    points = Visual.visual_info[str(Visual.cur_snap)]["POINTS"]
    point_ids = Visual.visual_info[str(Visual.cur_snap)]["POINT_IDS"]
    point_Group_ids = Visual.visual_info[str(Visual.cur_snap)]["POINT_GROUP_IDS"]

    lines_point_pair =Visual.visual_info[str(Visual.cur_snap)]["LINE_POINTS"]
    lines_id_map = Visual.visual_info[str(Visual.cur_snap)]["LINE_IDS"]

    Tplates_point_pair = Visual.visual_info[str(Visual.cur_snap)]["TRI_POINTS"]
    Tplates_id_map = Visual.visual_info[str(Visual.cur_snap)]["TRI_IDS"]

    Qplates_point_pair = Visual.visual_info[str(Visual.cur_snap)]["QUAD_POINTS"]
    Qplates_id_map =Visual.visual_info[str(Visual.cur_snap)]["QUAD_IDS"]



    if lines_point_pair!=[] and Visual.toggle_Beam:
        msh = pv.PolyData(points , lines=lines_point_pair)
        msh.cell_data["ids"] = lines_id_map
        plotter.add_mesh(msh, scalars="ids", cmap="plasma", line_width=4,show_edges=False,opacity=0.95,show_scalar_bar=False,name="Lines")

    if Qplates_point_pair!=[] and Visual.toggle_Plate:
        Qmesh = pv.PolyData(points, Qplates_point_pair)
        Qmesh.cell_data["ids"] = Qplates_id_map
        plotter.add_mesh(Qmesh, scalars="ids", cmap="rainbow", show_edges=True,opacity=0.8,edge_opacity=0.3,show_scalar_bar=False,name="QPlates")

    if Tplates_point_pair!=[] and Visual.toggle_Plate:
        Tmesh = pv.PolyData(points, Tplates_point_pair)
        Tmesh.cell_data["ids"] = Tplates_id_map
        plotter.add_mesh(Tmesh, scalars="ids", cmap="rainbow", show_edges=True,opacity=0.8,edge_opacity=0.3,show_scalar_bar=False,name="TPlates")






    # SHOW NODE ID ---------------------------------------------------------------------------------------
    if Visual.toggle_NodeIDs and points!=[]:
        plotter.add_point_labels(points, point_ids, 
                                    always_visible=True,shape=None,show_points=True,
                                    fill_shape=False,
                                    point_size=10,point_color="red",
                                    font_size=15, name="LABEL_ID"
                                    )

    
    #----------------------------------------------------------------------------------------------------
    
    #GROUP NODE ID ------------------------------------------------------------------------------------
    if Visual.toggle_GroupNodeIDs and points!=[]:
        plotter.add_point_labels(points, point_Group_ids, 
                                    always_visible=True,shape=None,show_points=True,
                                    fill_shape=False,
                                    point_size=10,point_color="orange",
                                    font_size=17, name="GLABEL_ID"
                                    )

    

   # -----------------------------------------------------------------------------------------------------



    # GRID ----------------------------------------------------------------------------------------------
    x_offset = 0.2*(max_x-min_x)+1
    y_offset = 0.2*(max_y-min_y)+1
    x = np.linspace(min_x-x_offset, max_x+x_offset, 21)
    y = np.linspace(min_y-y_offset, max_y+y_offset, 21)


    lines = []
    z_off = 1

    for xi in x:
        lines.append([[xi, y[0], min_z-z_off], [xi, y[-1], min_z-z_off]])  # vertical lines

    for yi in y:
        lines.append([[x[0], yi, min_z-z_off], [x[-1], yi, min_z-z_off]])  # horizontal lines


    grid_lines = pv.PolyData()
    for line in lines:
        grid_lines += pv.Line(line[0], line[1])

    
    if Visual.toggle_Grid:
        plotter.add_mesh(grid_lines, color='gray', line_width=1,name="Grid",opacity=0.3)


#--------------------------------------------------------------------------------------------------------------


    def toggle_GridLines(checked):
        if checked:
            Visual.toggle_Grid = True
            plotter.add_mesh(grid_lines, color='gray', line_width=1,name="Grid",opacity=0.3)
        else:
            Visual.toggle_Grid = False
            plotter.remove_actor("Grid")
        plotter.render()


    def toggle_GroupNodeID(checked):
        if checked:
            Visual.toggle_GroupNodeIDs = True
            plotter.add_point_labels(points, point_Group_ids, 
                                always_visible=True,shape=None,show_points=True,
                                fill_shape=False,
                                point_size=10,point_color="orange",
                                font_size=17, name="GLABEL_ID"
                                )
        else:
            Visual.toggle_GroupNodeIDs = False
            plotter.remove_actor("GLABEL_ID")        
        
        plotter.render()



    def toggle_Beams(checked):
        if checked:
            Visual.toggle_Beam = True
            if lines_point_pair!=[]: plotter.add_mesh(msh, scalars="ids", cmap="plasma", line_width=4,show_edges=False,opacity=0.95,show_scalar_bar=False,name="Lines")
        else:
            Visual.toggle_Beam = False
            plotter.remove_actor("Lines")
        plotter.render()

    def toggle_Plates(checked):
        if checked:
            Visual.toggle_Plate = True
            if Qplates_point_pair!=[]:plotter.add_mesh(Qmesh, scalars="ids", cmap="rainbow", show_edges=True,opacity=0.8,edge_opacity=0.3,show_scalar_bar=False,name="QPlates")
            if Tplates_point_pair!=[]:plotter.add_mesh(Tmesh, scalars="ids", cmap="rainbow", show_edges=True,opacity=0.8,edge_opacity=0.3,show_scalar_bar=False,name="TPlates")
        else:
            Visual.toggle_Plate = False
            plotter.remove_actor("QPlates")
            plotter.remove_actor("TPlates")
        plotter.render()

    def toggle_nodeID(checked):
        if checked:
            Visual.toggle_NodeIDs = True
            plotter.add_point_labels(points, point_ids, 
                                always_visible=True,shape=None,show_points=True,
                                fill_shape=False,
                                point_size=10,point_color="red",
                                font_size=15, name="LABEL_ID"
                                )
        else:
            Visual.toggle_NodeIDs = False
            plotter.remove_actor("LABEL_ID")

        plotter.render()

    
    cbox_NodeID_pos = (50,10)
    plotter.add_checkbox_button_widget(toggle_nodeID, value=Visual.toggle_NodeIDs, position=cbox_NodeID_pos, size=20)
    plotter.add_text("Node ID", font_size=6, color='black',position=(cbox_NodeID_pos[0]+25,cbox_NodeID_pos[1]+3))

    cbox_GrupNodeID_pos = (150,10)
    plotter.add_checkbox_button_widget(toggle_GroupNodeID, value=Visual.toggle_GroupNodeIDs, position=cbox_GrupNodeID_pos, size=20)
    plotter.add_text("Group Node ID", font_size=6, color='black',position=(cbox_GrupNodeID_pos[0]+25,cbox_GrupNodeID_pos[1]+3))

    cbox_DispBeam_pos = (280,10)
    plotter.add_checkbox_button_widget(toggle_Beams, value=Visual.toggle_Beam, position=cbox_DispBeam_pos, size=20)
    plotter.add_text("Beams", font_size=6, color='black',position=(cbox_DispBeam_pos[0]+25,cbox_DispBeam_pos[1]+3))

    cbox_DispPlate_pos = (370,10)
    plotter.add_checkbox_button_widget(toggle_Plates, value=Visual.toggle_Plate, position=cbox_DispPlate_pos, size=20)
    plotter.add_text("Plates", font_size=6, color='black',position=(cbox_DispPlate_pos[0]+25,cbox_DispPlate_pos[1]+3))

    cbox_GridDispl_pos = (450,10)
    plotter.add_checkbox_button_widget(toggle_GridLines, value=Visual.toggle_Grid, position=cbox_GridDispl_pos, size=20)
    plotter.add_text("Grid", font_size=6, color='black',position=(cbox_GridDispl_pos[0]+25,cbox_GridDispl_pos[1]+3))

    cbox_Animate_pos = (580,10)
    plotter.add_checkbox_button_widget(changeDataBack, value=True, position=cbox_Animate_pos, size=20)
    plotter.add_text("<<", font_size=6, color='red',position=(cbox_Animate_pos[0]+25,cbox_Animate_pos[1]+3))

    cbox_Animate_pos = (640,10)
    plotter.add_checkbox_button_widget(changeDataForw, value=True, position=cbox_Animate_pos, size=20)
    plotter.add_text(">>", font_size=6, color='green',position=(cbox_Animate_pos[0]+25,cbox_Animate_pos[1]+3))

    
    #--------------------------------------------------------------------------------------------------------------


    return plotter



from ._mapi import MidasAPI

class CS:

    @staticmethod
    def create():
        if CS.STAGE.stages!=[]: CS.STAGE.create()
        if CS.CompSec.compsecs!=[] : CS.CompSec.create()
        if CS.TimeLoad.timeloads!=[] : CS.TimeLoad.create()
        if CS.CreepCoeff.creepcoeffs!=[] : CS.CreepCoeff.create()
        if CS.Camber.cambers!=[] : CS.Camber.create()

    class STAGE:
        stages = []
        _maxID_ = 0
        _maxNO_ = 0
        _isSync_ = False

        def __init__(self, 
                    name: str,
                    duration: float = 0, 
                    s_group: str = None, 
                    s_age: float = None, 
                    s_type: str= None, 
                    b_group: str = None, 
                    b_pos: str = None, 
                    b_type: str = None,
                    l_group: str = None, 
                    l_day: str = None, 
                    l_type: str = None, 
                    id: int = None, 
                    sv_result: bool = True, 
                    sv_step: bool = False, 
                    load_in: bool = False, 
                    nl: int = 5, 
                    addstp: list = None):
            """
            Construction Stage define.
            
            Parameters:
                name: Name of Construction Stage
                duration: Duration of Construction Stage in days (default 0)
                s_group: Structure group name or list of group names (default None)
                s_age: Age of structure group in days and Redistribution value(%) win case of Deactivation (default 0)
                s_type: Structure activation type - "A" to activate, "D" to deactivate(default A)
                b_group: Boundary group name or list of group names (default None)
                b_pos: Boundary position type - "ORIGINAL" or "DEFORMED", or list (default DEFORMED)
                b_type: Boundary activation type - "A" to activate, "D" to deactivate (default A)
                l_group: Load group name or list of group names (default None)
                l_day: Load activation day - "FIRST" or "LAST" (default "FIRST")
                l_type: Load activation type - "A" to activate, "D" to deactivate (default A)
                id: The construction stage ID (optional)
                sv_result: Save results of this stage (default True)
                sv_step: Add additional step results (default False)
                load_in: Load incremental steps for material nonlinear analysis (default False)
                nl: Number of load incremental steps (default 5)
                addstp: List of additional steps (default None)
            
            Examples:
                ```python
                # Single group activation
                CS("CS1", 7, "S1", 7, "A", "B1", "DEFORMED", "A", "L1", "FIRST", "A")
                
                # Multiple group activation
                CS("CS1", 7, ["S1", "S2"], [7, 10], ["A", "A"], ["B1", "B2"], 
                ["DEFORMED", "DEFORMED"], ["A", "A"], ["L1", "L2"], ["FIRST", "FIRST"], ["A", "A"])
                
                # Mixed activation and deactivation
                CS("CS1", 7, ["S1", "S2"], [7, 10], ["A", "D"], ["B1", "B2"], 
                ["DEFORMED", "DEFORMED"], ["A", "D"], "L1", "FIRST", "A")
                
                # With additional options
                CS("CS1", 7, "S1", 7, "A", "B1", "DEFORMED", "A", "L1", "FIRST", "A",
                sv_result=True, sv_step=True, load_in=True, nl=6, addstp=[1, 2, 3])
                ```
            """

            self.NAME = name
            self.DURATION = duration
            self.SV_Result = sv_result
            self.SV_Step = sv_step
            self.Load_IN = load_in
            self.NL = nl
            self.addstp = [] if addstp is None else addstp
            
            # Initialize group containers
            self.act_structure_groups = []  
            self.deact_structure_groups = []  
            self.act_boundary_groups = []  
            self.deact_boundary_groups = []  
            self.act_load_groups = []  
            self.deact_load_groups = []  
            

            # Set ID
            if id is None:
                self.ID = CS.STAGE._maxID_ + 1
                self.NO = CS.STAGE._maxNO_ + 1
            else:
                self.ID = id
                self.NO = id
            CS.STAGE._maxNO_ = max(CS.STAGE._maxNO_ ,self.NO)
            CS.STAGE._maxID_ = max(CS.STAGE._maxID_ ,self.ID)

            # Process structure groups
            if s_group:
                # Convert single values to lists for uniform processing
                if not isinstance(s_group, list):
                    s_group = [s_group]
                    s_age = [s_age if s_age is not None else 0]
                    s_type = [s_type if s_type is not None else "A"]
                else:
                    # Ensure other parameters are lists too
                    if s_age is None:
                        s_age = [0] * len(s_group)
                    elif not isinstance(s_age, list):
                        s_age = [s_age] * len(s_group)
                    
                    if s_type is None:
                        s_type = ["A"] * len(s_group)
                    elif not isinstance(s_type, list):
                        s_type = [s_type] * len(s_group)
                
                # Process each structure group
                for i, group in enumerate(s_group):
                    if i < len(s_type) and s_type[i] == "A":
                        # Activation: Check if already activated in previous stages
                        for stage in CS.STAGE.stages:
                            for existing_group in stage.act_structure_groups:
                                if existing_group["name"] == group:
                                    raise ValueError(f"Structure group '{group}' has already been activated in stage '{stage.NAME}' (ID: {stage.ID})")
                        
                        age = s_age[i] if i < len(s_age) else 0
                        self.act_structure_groups.append({"name": group, "age": age})
                    else:
                        # Deactivation: Check if activated in previous stages
                        activated = False
                        for stage in CS.STAGE.stages:
                            for existing_group in stage.act_structure_groups:
                                if existing_group["name"] == group:
                                    activated = True
                                    break
                            if activated:
                                break
                        
                        if not activated:
                            raise ValueError(f"Structure group '{group}' cannot be deactivated as it has not been activated in any previous stage")
                        
                        # For deactivation, s_age value is used as redist percentage
                        redist = s_age[i] if i < len(s_age) else 100
                        self.deact_structure_groups.append({"name": group, "redist": redist})
            
            # Process boundary groups
            if b_group:
                # Convert single values to lists for uniform processing
                if not isinstance(b_group, list):
                    b_group = [b_group]
                    b_pos = [b_pos if b_pos is not None else "DEFORMED"]
                    b_type = [b_type if b_type is not None else "A"]
                else:
                    # Ensure other parameters are lists too
                    if b_pos is None:
                        b_pos = ["DEFORMED"] * len(b_group)
                    elif not isinstance(b_pos, list):
                        b_pos = [b_pos] * len(b_group)
                    
                    if b_type is None:
                        b_type = ["A"] * len(b_group)
                    elif not isinstance(b_type, list):
                        b_type = [b_type] * len(b_group)
                
                # Process each boundary group
                for i, group in enumerate(b_group):
                    if i < len(b_type) and b_type[i] == "A":
                        # Activation: Check if already activated in previous stages
                        for stage in CS.STAGE.stages:
                            for existing_group in stage.act_boundary_groups:
                                if existing_group["name"] == group:
                                    raise ValueError(f"Boundary group '{group}' has already been activated in stage '{stage.NAME}' (ID: {stage.ID})")
                        
                        pos = b_pos[i] if i < len(b_pos) else "DEFORMED"
                        self.act_boundary_groups.append({"name": group, "pos": pos})
                    else:
                        # Deactivation: Check if activated in previous stages
                        activated = False
                        for stage in CS.STAGE.stages:
                            for existing_group in stage.act_boundary_groups:
                                if existing_group["name"] == group:
                                    activated = True
                                    break
                            if activated:
                                break
                        
                        if not activated:
                            raise ValueError(f"Boundary group '{group}' cannot be deactivated as it has not been activated in any previous stage")
                        
                        self.deact_boundary_groups.append(group)
            
            # Process load groups
            if l_group:
                # Convert single values to lists for uniform processing
                if not isinstance(l_group, list):
                    l_group = [l_group]
                    l_day = [l_day if l_day is not None else "FIRST"]
                    l_type = [l_type if l_type is not None else "A"]
                else:
                    # Ensure other parameters are lists too
                    if l_day is None:
                        l_day = ["FIRST"] * len(l_group)
                    elif not isinstance(l_day, list):
                        l_day = [l_day] * len(l_group)
                    
                    if l_type is None:
                        l_type = ["A"] * len(l_group)
                    elif not isinstance(l_type, list):
                        l_type = [l_type] * len(l_group)
                
                # Process each load group
                for i, group in enumerate(l_group):
                    if i < len(l_type) and l_type[i] == "A":
                        # Activation: Check if already activated in previous stages
                        for stage in CS.STAGE.stages:
                            for existing_group in stage.act_load_groups:
                                if existing_group["name"] == group:
                                    raise ValueError(f"Load group '{group}' has already been activated in stage '{stage.NAME}' (ID: {stage.ID})")
                        
                        day = l_day[i] if i < len(l_day) else "FIRST"
                        self.act_load_groups.append({"name": group, "day": day})
                    else:
                        # Deactivation: Check if activated in previous stages
                        activated = False
                        for stage in CS.STAGE.stages:
                            for existing_group in stage.act_load_groups:
                                if existing_group["name"] == group:
                                    activated = True
                                    break
                            if activated:
                                break
                        
                        if not activated:
                            raise ValueError(f"Load group '{group}' cannot be deactivated as it has not been activated in any previous stage")
                        
                        day = l_day[i] if i < len(l_day) else "FIRST"
                        self.deact_load_groups.append({"name": group, "day": day})
            
            CS.STAGE.stages.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts Construction Stage data to JSON format 
            Example:
                # Get the JSON data for all construction stages
                json_data = CS.json()
                print(json_data)
            """
            json = {"Assign": {}}
            
            for csa in cls.stages:
                stage_data = {
                    "NAME": csa.NAME,
                    "DURATION": csa.DURATION,
                    "bSV_RSLT": csa.SV_Result,
                    "bSV_STEP": csa.SV_Step,
                    "bLOAD_STEP": csa.Load_IN,
                    "NO" : csa.NO
                }
                
                # Add incremental steps if load step is enabled
                if csa.Load_IN:
                    stage_data["INCRE_STEP"] = csa.NL
                
                # Add additional steps if specified
                if csa.addstp:
                    stage_data["ADD_STEP"] = csa.addstp
                else:
                    stage_data["ADD_STEP"] = []
                
                # Handle structure group activation
                if csa.act_structure_groups:
                    stage_data["ACT_ELEM"] = []
                    for group in csa.act_structure_groups:
                        stage_data["ACT_ELEM"].append({
                            "GRUP_NAME": group["name"],
                            "AGE": group["age"]
                        })
                
                # Handle structure group deactivation
                if csa.deact_structure_groups:
                    stage_data["DACT_ELEM"] = []
                    for group in csa.deact_structure_groups:
                        stage_data["DACT_ELEM"].append({
                            "GRUP_NAME": group["name"],
                            "REDIST": group["redist"]
                        })
                
                # Handle boundary group activation
                if csa.act_boundary_groups:
                    stage_data["ACT_BNGR"] = []
                    for group in csa.act_boundary_groups:
                        stage_data["ACT_BNGR"].append({
                            "BNGR_NAME": group["name"],
                            "POS": group["pos"]
                        })
                
                # Handle boundary group deactivation
                if csa.deact_boundary_groups:
                    stage_data["DACT_BNGR"] = []
                    for group_name in csa.deact_boundary_groups:
                        stage_data["DACT_BNGR"].append(group_name)
                
                # Handle load group activation
                if csa.act_load_groups:
                    stage_data["ACT_LOAD"] = []
                    for group in csa.act_load_groups:
                        stage_data["ACT_LOAD"].append({
                            "LOAD_NAME": group["name"],
                            "DAY": group["day"]
                        })
                
                # Handle load group deactivation
                if csa.deact_load_groups:
                    stage_data["DACT_LOAD"] = []
                    for group in csa.deact_load_groups:
                        stage_data["DACT_LOAD"].append({
                            "LOAD_NAME": group["name"],
                            "DAY": group["day"]
                        })
                
                json["Assign"][str(csa.ID)] = stage_data
            
            return json
        
        @classmethod
        def create(cls):
            """Creates construction stages in the database"""
            if CS.STAGE._isSync_:
                MidasAPI("DELETE", "/db/stag")
            MidasAPI("PUT", "/db/stag", cls.json())
        
        @classmethod
        def get(cls):
            """Gets construction stage data from the database"""
            return MidasAPI("GET", "/db/stag")
        
        @classmethod
        def sync(cls):
            """Updates the CS class with data from the database"""
            cls.stages = []
            a = cls.get()
            if a != {'message': ''}:
                if "STAG" in a:
                    stag_data_dict = a["STAG"]
                else:
                    return  
                    
                for stag_id, stag_data in stag_data_dict.items():
                    # Basic stage data
                    name = stag_data.get("NAME")
                    duration = stag_data.get("DURATION")
                    sv_result = stag_data.get("bSV_RSLT")
                    sv_step = stag_data.get("bSV_STEP")
                    load_in = stag_data.get("bLOAD_STEP")
                    nl = stag_data.get("INCRE_STEP")
                    addstp = stag_data.get("ADD_STEP")
                    stagNo = stag_data.get("NO")
                    
                    # Create a new CS object with basic properties
                    new_cs = CS.STAGE(
                        name=name,
                        duration=duration,
                        id=int(stagNo),
                        sv_result=sv_result,
                        sv_step=sv_step,
                        load_in=load_in,
                        nl=nl,
                        addstp=addstp
                    )
                    new_cs.NO = stagNo
                    CS.STAGE.stages.pop()
                    
                    # Process activation elements
                    if "ACT_ELEM" in stag_data and stag_data["ACT_ELEM"]:
                        for elem in stag_data["ACT_ELEM"]:
                            group_name = elem.get("GRUP_NAME")
                            age = elem.get("AGE")
                            new_cs.act_structure_groups.append({"name": group_name, "age": age})
                    
                    # Process deactivation elements
                    if "DACT_ELEM" in stag_data and stag_data["DACT_ELEM"]:
                        for elem in stag_data["DACT_ELEM"]:
                            if isinstance(elem, dict):
                                group_name = elem.get("GRUP_NAME")
                                redist = elem.get("REDIST")
                            else:
                                group_name = elem
                                redist = 0
                            new_cs.deact_structure_groups.append({"name": group_name, "redist": redist})
                    
                    # Process activation boundary groups
                    if "ACT_BNGR" in stag_data and stag_data["ACT_BNGR"]:
                        for bngr in stag_data["ACT_BNGR"]:
                            group_name = bngr.get("BNGR_NAME")
                            pos = bngr.get("POS")
                            new_cs.act_boundary_groups.append({"name": group_name, "pos": pos})
                    
                    # Process deactivation boundary groups
                    if "DACT_BNGR" in stag_data and stag_data["DACT_BNGR"]:
                        for bngr in stag_data["DACT_BNGR"]:
                            new_cs.deact_boundary_groups.append(bngr)
                    
                    # Process activation loads
                    if "ACT_LOAD" in stag_data and stag_data["ACT_LOAD"]:
                        for load in stag_data["ACT_LOAD"]:
                            group_name = load.get("LOAD_NAME")
                            day = load.get("DAY")
                            new_cs.act_load_groups.append({"name": group_name, "day": day})
                    
                    # Process deactivation loads
                    if "DACT_LOAD" in stag_data and stag_data["DACT_LOAD"]:
                        for load in stag_data["DACT_LOAD"]:
                            if isinstance(load, dict):
                                group_name = load.get("LOAD_NAME")
                                day = load.get("DAY")
                            else:
                                group_name = load
                                day = "FIRST"
                            new_cs.deact_load_groups.append({"name": group_name, "day": day})
                    
                    CS.STAGE.stages.append(new_cs)

                sorted_stgs = sorted(CS.STAGE.stages,key=lambda x : x.NO)
                CS.STAGE.stages = sorted_stgs
                CS.STAGE._isSync_ = True
        
        @classmethod
        def delete(cls):
            """Deletes all construction stages from the database and resets the class"""
            cls.stages = []
            return MidasAPI("DELETE", "/db/stag")
        
#-----------------------------------------------------------Comp Section for CS--------------------------------------------------------------

    class CompSec:
        compsecs = []

        def __init__(self, 
                    activation_stage: str,
                    section_id: int,
                    comp_type: str = "GENERAL",
                    tapered_type: bool = False,
                    partinfo: list = None,
                    id: int = None):
            """
            Parameters:
                activation_stage: Active Stage name (required)
                section_id: Section ID (required)
                comp_type: Composite Type - "GENERAL" or "USER" (default "GENERAL")
                tapered_type: Tapered Type - True or False (default False)
                partinfo: List of part information lists (required)
                id: The composite section ID (optional)
            
            Part Info Format:
                Each part should be a list with elements in order:
                [part_number, material_type, material_id, composite_stage, age, 
                 height, volume_surface_ratio, module_exposed_surface, area, 
                 asy, asz, ixx, iyy, izz, warea, iw]
                
                - part_number: Integer (required)
                - material_type: "ELEM" or "MATL" (required)
                - material_id: String (optional, blank for ELEM)
                - composite_stage: String (optional, blank for active stage)
                - age: Number (default 0)
                - height: Number (default AUTO)
                - volume_surface_ratio: Number (default 0)
                - module_exposed_surface: Number (default 0)
                - area: Number (default 1)
                - asy: Number (default 1)
                - asz: Number (default 1)
                - ixx: Number (default 1)
                - iyy: Number (default 1)
                - izz: Number (default 1)
                - warea: Number (default 1)
                - iw: Number (default 1)
            
            Examples:
                ```python
                # Basic composite section
                CompSec("CS1", 1, "GENERAL", False, [
                    [1, "ELEM", "", "", 2, 1.5, 1.5, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                    [2, "MATL", "3", "CS2", 5, 0.245, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]
                ])
                
                # With minimal part info (using defaults)
                CompSec("CS1", 2, "GENERAL", False, [
                    [1, "ELEM"],
                    [2, "MATL", "2","CS2"]
                ])
                ```
            """
            
            self.ASTAGE = activation_stage
            self.SEC = section_id
            self.TYPE = comp_type
            self.bTAP = tapered_type
            
            # Set ID
            if id is None:
                self.ID = len(CS.CompSec.compsecs) + 1
            else:
                self.ID = id
            
            # Process part information
            self.vPARTINFO = []
            
            if partinfo is None:
                raise ValueError("Part information is required")
            
            if not isinstance(partinfo, list):
                raise ValueError("Part information must be a list of lists")
            
            for part_data in partinfo:
                if not isinstance(part_data, list) or len(part_data) < 2:
                    raise ValueError("Each part must be a list with at least part number and material type")
                
                # Default values for part info
                defaults = [
                    None,  # PART (required)
                    None,  # MTYPE (required)
                    "",    # MAT
                    "",    # CSTAGE
                    0,     # AGE
                    "AUTO", # PARTINFO_H
                    0,     # PARTINFO_VS
                    0,     # PARTINFO_M
                    1,     # AREA
                    1,     # ASY
                    1,     # ASZ
                    1,     # IXX
                    1,     # IYY
                    1,     # IZZ
                    1,     # WAREA
                    1      # IW
                ]
                
                # Fill in provided values
                for i, value in enumerate(part_data):
                    if i < len(defaults):
                        defaults[i] = value
                
                # Validate required fields
                if defaults[0] is None:
                    raise ValueError("Part number is required")
                if defaults[1] is None:
                    raise ValueError("Material type is required")
                if defaults[1] not in ["ELEM", "MATL"]:
                    raise ValueError("Material type must be 'ELEM' or 'MATL'")
                
                # Create part info dictionary
                part_info = {
                    "PART": defaults[0],
                    "MTYPE": defaults[1],
                    "MAT": defaults[2],
                    "CSTAGE": defaults[3],
                    "AGE": defaults[4],
                    "PARTINFO_H": defaults[5],
                    "PARTINFO_VS": defaults[6],
                    "PARTINFO_M": defaults[7],
                    "AREA": defaults[8],
                    "ASY": defaults[9],
                    "ASZ": defaults[10],
                    "IXX": defaults[11],
                    "IYY": defaults[12],
                    "IZZ": defaults[13],
                    "WAREA": defaults[14],
                    "IW": defaults[15]
                }
                
                self.vPARTINFO.append(part_info)
            
            CS.CompSec.compsecs.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts Composite Section data to JSON format 
            Example:
                # Get the JSON data for all composite sections
                json_data = CS.CompSec.json()
                print(json_data)
            """
            json_data = {"Assign": {}}
            
            for compsec in cls.compsecs:
                section_data = {
                    "SEC": compsec.SEC,
                    "ASTAGE": compsec.ASTAGE,
                    "TYPE": compsec.TYPE,
                    "bTAP": compsec.bTAP,
                    "vPARTINFO": compsec.vPARTINFO
                }
                
                json_data["Assign"][str(compsec.ID)] = section_data
            
            return json_data
        
        @classmethod
        def create(cls):
            """Creates composite sections in the database"""
            return MidasAPI("PUT", "/db/cscs", cls.json())
        
        @classmethod
        def get(cls):
            """Gets composite section data from the database"""
            return MidasAPI("GET", "/db/cscs")
        
        @classmethod
        def sync(cls):
            """Updates the CompSec class with data from the database"""
            cls.compsecs = []
            response = cls.get()
            
            if response != {'message': ''}:
                if "CSCS" in response:
                    cscs_data_dict = response["CSCS"]
                else:
                    return
                
                for cscs_id, cscs_data in cscs_data_dict.items():
                    # Basic section data
                    astage = cscs_data.get("ASTAGE")
                    sec = cscs_data.get("SEC")
                    comp_type = cscs_data.get("TYPE", "GENERAL")
                    tapered_type = cscs_data.get("bTAP", False)
                    partinfo_data = cscs_data.get("vPARTINFO", [])
                    
                    # Convert partinfo from dict format to list format
                    partinfo = []
                    for part in partinfo_data:
                        part_list = [
                            part.get("PART"),
                            part.get("MTYPE"),
                            part.get("MAT", ""),
                            part.get("CSTAGE", ""),
                            part.get("AGE", 0),
                            part.get("PARTINFO_H", "AUTO"),
                            part.get("PARTINFO_VS", 0),
                            part.get("PARTINFO_M", 0),
                            part.get("AREA", 1),
                            part.get("ASY", 1),
                            part.get("ASZ", 1),
                            part.get("IXX", 1),
                            part.get("IYY", 1),
                            part.get("IZZ", 1),
                            part.get("WAREA", 1),
                            part.get("IW", 1)
                        ]
                        partinfo.append(part_list)
                    
                    # Create a new CompSec object
                    new_compsec = CS.CompSec(
                        activation_stage=astage,
                        section_id=sec,
                        comp_type=comp_type,
                        tapered_type=tapered_type,
                        partinfo=partinfo,
                        id=int(cscs_id)
                    )
                    
                    # Remove the automatically added instance and replace with synced data
                    CS.CompSec.compsecs.pop()
                    CS.CompSec.compsecs.append(new_compsec)
        
        @classmethod
        def delete(cls):
            """Deletes all composite sections from the database and resets the class"""
            cls.compsecs = []
            return MidasAPI("DELETE", "/db/cscs")


#-----------------------------------------------------------------------------------------------------------------------------------

    class TimeLoad:
        timeloads = []

        def __init__(self, 
                    element_id: int,
                    day: int,
                    group: str = "",
                    id: int = None):
            """
            Time Loads for Construction Stage define.
            
            Parameters:
                element_id: Element ID (required)
                day: Time Loads in days (required)
                group: Load Group Name (optional, default blank)
                id: The time loads ID (optional)
            
            Examples:
                ```python
                CS.TimeLoad(10, 35, "DL")
                ```
            """
            
            self.ELEMENT_ID = element_id
            self.DAY = day
            self.GROUP_NAME = group
            
            # Set ID
            if id is None:
                self.ID = len(CS.TimeLoad.timeloads) + 1
            else:
                self.ID = id
            
            CS.TimeLoad.timeloads.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts Time Loads data to JSON format 
            Example:
                # Get the JSON data for all time loads
                json_data = CS.TimeLoad.json()
                print(json_data)
            """
            json_data = {"Assign": {}}
            
            for timeload in cls.timeloads:
                items_data = {
                    "ITEMS": [
                        {
                            "ID": 1,
                            "GROUP_NAME": timeload.GROUP_NAME,
                            "DAY": timeload.DAY
                        }
                    ]
                }
                
                json_data["Assign"][str(timeload.ELEMENT_ID)] = items_data
            
            return json_data
        
        @classmethod
        def create(cls):
            """Creates time loads in the CIVIL NX"""
            return MidasAPI("PUT", "/db/tmld", cls.json())
        
        @classmethod
        def get(cls):
            """Gets time loads data from the CIVIL NX"""
            return MidasAPI("GET", "/db/tmld")
        
        @classmethod
        def sync(cls):
            """Updates the TimeLoad class with data from the CIVIL NX"""
            cls.timeloads = []
            response = cls.get()
            
            if response != {'message': ''}:
                if "TMLD" in response:
                    stbk_data_dict = response["TMLD"]
                else:
                    return
                
                for element_id, stbk_data in stbk_data_dict.items():
                    items = stbk_data.get("ITEMS", [])
                    
                    for item in items:
                        group_name = item.get("GROUP_NAME", "")
                        day = item.get("DAY", 0)
                        item_id = item.get("ID", 1)
                        
                        # Create a new TimeLoad object
                        new_timeload = CS.TimeLoad(
                            element_id=int(element_id),
                            day=day,
                            group=group_name,
                            id=item_id
                        )
                        
                        # Remove the automatically added instance and replace with synced data
                        CS.TimeLoad.timeloads.pop()
                        CS.TimeLoad.timeloads.append(new_timeload)
        
        @classmethod
        def delete(cls):
            """Deletes all time loads from the CIVIL NX and python class"""
            cls.timeloads = []
            return MidasAPI("DELETE", "/db/tmld")

    class CreepCoeff:
        creepcoeffs = []

        def __init__(self, 
                    element_id: int,
                    creep: float,
                    group: str = "",
                    id: int = None):
            """
            Creep Coefficient for Construction Stage define.
            
            Parameters:
                element_id: Element ID (required)
                creep: Creep Coefficient value (required)
                group: Load Group Name (optional, default blank)
                id: The creep coefficient ID (optional)
            
            Examples:
                ```python
                # Basic creep coefficient
                CS.CreepCoeff(25, 1.2)   

                # With specific ID & Group
                CS.CreepCoeff(26, 1.5, "GR", id=2)
                ```
            """
            
            self.ELEMENT_ID = element_id
            self.CREEP = creep
            self.GROUP_NAME = group
            
            # Set ID
            if id is None:
                self.ID = len(CS.CreepCoeff.creepcoeffs) + 1
            else:
                self.ID = id
            
            CS.CreepCoeff.creepcoeffs.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts Creep Coefficient data to JSON format 
            Example:
                # Get the JSON data for all creep coefficients
                json_data = CS.CreepCoeff.json()
                print(json_data)
            """
            json_data = {"Assign": {}}
            
            for creepcoeff in cls.creepcoeffs:
                items_data = {
                    "ITEMS": [
                        {
                            "ID": 1,
                            "GROUP_NAME": creepcoeff.GROUP_NAME,
                            "CREEP": creepcoeff.CREEP
                        }
                    ]
                }
                
                json_data["Assign"][str(creepcoeff.ELEMENT_ID)] = items_data
            
            return json_data
        
        @classmethod
        def create(cls):
            """Creates creep coefficients in the database"""
            return MidasAPI("PUT", "/db/crpc", cls.json())
        
        @classmethod
        def get(cls):
            """Gets creep coefficient data from the database"""
            return MidasAPI("GET", "/db/crpc")
        
        @classmethod
        def sync(cls):
            """Updates the CreepCoeff class with data from the database"""
            cls.creepcoeffs = []
            response = cls.get()
            
            if response != {'message': ''}:
                if "CRPC" in response:
                    crpc_data_dict = response["CRPC"]
                else:
                    return
                
                for element_id, crpc_data in crpc_data_dict.items():
                    items = crpc_data.get("ITEMS", [])
                    
                    for item in items:
                        group_name = item.get("GROUP_NAME", "")
                        creep = item.get("CREEP", 0.0)
                        item_id = item.get("ID", 1)
                        
                        # Create a new CreepCoeff object
                        new_creepcoeff = CS.CreepCoeff(
                            element_id=int(element_id),
                            creep=creep,
                            group=group_name,
                            id=item_id
                        )
                        
                        # Remove the automatically added instance and replace with synced data
                        CS.CreepCoeff.creepcoeffs.pop()
                        CS.CreepCoeff.creepcoeffs.append(new_creepcoeff)
        
        @classmethod
        def delete(cls):
            """Deletes all creep coefficients from the database and resets the class"""
            cls.creepcoeffs = []
            return MidasAPI("DELETE", "/db/crpc")

    class Camber:
        cambers = []

        def __init__(self, 
                    node_id: int,
                    camber: float,
                    deform: float,
                    id: int = None):
            """
            Camber for Construction Stage define.
            
            Parameters:
                node_id: Node ID (required)
                camber: User camber value (required)
                deform: Deformation value (required)
                id: The camber ID (optional)
            
            Examples:
                ```python
                
                CS.Camber(25, 0.17, 0.1)
                ```
            """
            
            self.NODE_ID = node_id
            self.USER = camber
            self.DEFORM = deform
            
            # Set ID
            if id is None:
                self.ID = len(CS.Camber.cambers) + 1
            else:
                self.ID = id
            
            CS.Camber.cambers.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts Camber data to JSON format 
            Example:
                # Get the JSON data for all cambers
                json_data = CS.Camber.json()
                print(json_data)
            """
            json_data = {"Assign": {}}
            
            for camber in cls.cambers:
                camber_data = {
                    "DEFORM": camber.DEFORM,
                    "USER": camber.USER
                }
                
                json_data["Assign"][str(camber.NODE_ID)] = camber_data
            
            return json_data
        
        @classmethod
        def create(cls):
            """Creates cambers in the database"""
            return MidasAPI("PUT", "/db/cmcs", cls.json())
        
        @classmethod
        def get(cls):
            """Gets camber data from the database"""
            return MidasAPI("GET", "/db/cmcs")
        
        @classmethod
        def sync(cls):
            """Updates the Camber class with data from the database"""
            cls.cambers = []
            response = cls.get()
            
            if response != {'message': ''}:
                if "CMCS" in response:
                    cmcs_data_dict = response["CMCS"]
                else:
                    return
                
                for node_id, cmcs_data in cmcs_data_dict.items():
                    deform = cmcs_data.get("DEFORM", 0.0)
                    user = cmcs_data.get("USER", 0.0)
                    
                    # Create a new Camber object
                    new_camber = CS.Camber(
                        node_id=int(node_id),
                        camber=user,
                        deform=deform,
                        id=len(cls.cambers) + 1
                    )
                    
                    # Remove the automatically added instance and replace with synced data
                    CS.Camber.cambers.pop()
                    CS.Camber.cambers.append(new_camber)
        
        @classmethod
        def delete(cls):
            """Deletes all cambers from the database and resets the class"""
            cls.cambers = []
            return MidasAPI("DELETE", "/db/cmcs")



import polars as pl
from ._mapi import *
from ._model import *

from ._mapi import _getUNIT
from ._mapi import _setUNIT
# js_file = open('JSON_Excel Parsing\\test.json','r')

# print(js_file)
# js_json = json.load(js_file)


#---- INPUT: JSON -> OUTPUT : Data FRAME --------- ---------
def _JSToDF_ResTable(js_json):
    res_json = {}

    c=0
    for heading in js_json["SS_Table"]["HEAD"]:
        for dat in js_json["SS_Table"]["DATA"]:
            try:
                res_json[heading].append(dat[c])
            except:
                res_json[heading]=[]
                res_json[heading].append(dat[c])

        c+=1

    res_df = pl.DataFrame(res_json)
    return(res_df)


def _Head_Data_2_DF_JSON(head,data):
    res_json = {}
    c=0
    for heading in head:
        for dat in data:
            try:
                res_json[heading].append(dat[c])
            except:
                res_json[heading]=[]
                res_json[heading].append(dat[c])

        c+=1
    return res_json
    

def _JSToDF_UserDefined(tableName,js_json,summary):

    if 'message' in js_json:
        print(f'⚠️  {tableName} table name does not exist.')
        Result.UserDefinedTables_print()
        return 'Check table name'
    
    if summary == 0:
        head = js_json[tableName]["HEAD"]
        data = js_json[tableName]["DATA"]
    elif summary > 0 :
        try :
            sub_tab1 = js_json[tableName]["SUB_TABLES"][summary-1]
            key_name = next(iter(sub_tab1))
            head = sub_tab1[key_name]["HEAD"]
            data = sub_tab1[key_name]["DATA"]
        except :
            print(' ⚠️  No Summary table exist')
            return 'No Summary table exist'


    res_json = _Head_Data_2_DF_JSON(head,data)
    res_df = pl.DataFrame(res_json)
    return(res_df)

    






# js_dat = {
#     "Argument": {
#         "TABLE_NAME": "SS_Table",
#         "TABLE_TYPE": "REACTIONG",
#         "UNIT": {
#             "FORCE": "kN",
#             "DIST": "m"
#         },
#         "STYLES": {
#             "FORMAT": "Fixed",
#             "PLACE": 12
#         }
#     }
# }

# MAPI_KEY('eyJ1ciI6InN1bWl0QG1pZGFzaXQuY29tIiwicGciOiJjaXZpbCIsImNuIjoib3R3aXF0NHNRdyJ9.da8f9dd41fee01425d8859e0091d3a46b0f252ff38341c46c73b26252a81571d')
# ss_json = MidasAPI("POST","/post/table",js_dat)
# df4 = JSToDF(ss_json)








# print(df4)
# df4.write_excel("new.xlsx",
#                 "Plate Forces",
#                 header_format={"bold":True},
#                 autofit=True,
#                 autofilter=True,
#                 table_style="Table Style Light 8"
#                 )


# with xlsxwriter.Workbook("test2.xlsx") as Wb:
#     ws = Wb.add_worksheet()

#     df4.write_excel(Wb,"Sheet 1",table_style="Table Style Light 8",autofit=True)

#     df4.write_excel(Wb,"Sheet 1",table_style="Table Style Light 8",autofit=True,autofilter=False,position="A31",include_header=False)





class Result :

    # ---------- User defined TABLE (Dynamic Report Table) ------------------------------
    @staticmethod
    def UserDefinedTable(tableName:str, summary=0, force_unit='KN',len_unit='M'):
        js_dat = {
            "Argument": {
                "TABLE_NAME": tableName,
                "STYLES": {
                    "FORMAT": "Fixed",
                    "PLACE": 5
                }
            }
        }
        currUNIT = _getUNIT()
        Model.units(force=force_unit,length=len_unit)
        ss_json = MidasAPI("POST","/post/TABLE",js_dat)
        _setUNIT(currUNIT)
        return _JSToDF_UserDefined(tableName,ss_json,summary)
    
    # ---------- LIST ALL USER DEFINED TABLE ------------------------------
    @staticmethod
    def UserDefinedTables_print():
        ''' Print all the User defined table names '''
        ss_json = MidasAPI("GET","/db/UTBL",{})
        table_name =[]
        try:
            for id in ss_json['UTBL']:
                table_name.append(ss_json["UTBL"][id]['NAME'])
            
            print('Available user-defined tables in Civil NX are : ')
            print(*table_name,sep=' , ')
        except:
            print(' ⚠️  There are no user-defined tables in Civil NX')



    # ---------- Result TABLE (For ALL TABLES)------------------------------
    @staticmethod
    def ResultTable(tabletype:str,keys=[],loadcase:list=[],cs_stage=[],force_unit='KN',len_unit='M'):
        '''
            TableType : REACTIONG | REACTIONL | DISPLACEMENTG | DISPLACEMENTL | TRUSSFORCE | TRUSSSTRESS
            Keys : List{int} -> Element/ Node IDs  |  str -> Structure Group Name
            Loadcase : Loadcase/Combination name followed by type. eg. DeadLoad(ST)
        '''

        js_dat = {
            "Argument": {
                "TABLE_NAME": "SS_Table",
                "TABLE_TYPE": tabletype,
                "STYLES": {
                    "FORMAT": "Fixed",
                    "PLACE": 5
                }
            }
        }

        if cs_stage !=[]:
            if cs_stage == 'all' :
                js_dat["Argument"]['OPT_CS'] = True
            else:
                js_dat["Argument"]['OPT_CS'] = True
                js_dat["Argument"]['STAGE_STEP'] = cs_stage


        if isinstance(keys,list):
            if keys!=[]:
                js_dat["Argument"]['NODE_ELEMS'] = {"KEYS": keys}
        elif isinstance(keys,str):
            js_dat["Argument"]['NODE_ELEMS'] = {"STRUCTURE_GROUP_NAME": keys}


        if loadcase!=[]: js_dat["Argument"]['LOAD_CASE_NAMES'] = loadcase

        currUNIT = _getUNIT()
        Model.units(force=force_unit,length=len_unit)
        ss_json = MidasAPI("POST","/post/table",js_dat)
        _setUNIT(currUNIT)
        return _JSToDF_ResTable(ss_json)
    

    class TABLE :
        
        @staticmethod
        def BeamForce_VBM(keys=[],loadcase:list=[],items=['all'],parts=["PartI", "PartJ"],components=['all'],force_unit='KN',len_unit='M'):
            '''
                Keys : List{int} -> Element/ Node IDs  |  str -> Structure Group Name
                Loadcase : Loadcase/Combination name followed by type. eg. ["DeadLoad(ST)"]
                Items to display : [ "Axial" , "Shear-y" , "Shear-z" , "Torsion" , "Moment-y" , "Moment-z"]
                Parts : ["PartI", "Part1/4", "Part2/4", "Part3/4", "PartJ"]
                Components (colms of tabulart result): [ "Elem", "Load", "Part", "Component", "Axial", "Shear-y", "Shear-z", "Torsion", "Moment-y", "Moment-z" ]
                
            '''

            js_dat = {
                "Argument": {
                    "TABLE_NAME": "SS_Table",
                    "TABLE_TYPE": "BEAMFORCEVBM",
                    "STYLES": {
                        "FORMAT": "Fixed",
                        "PLACE": 5
                    },
                    "PARTS" : parts
                }
            }


            if isinstance(keys,list):
                if keys!=[]:
                    js_dat["Argument"]['NODE_ELEMS'] = {"KEYS": keys}
            elif isinstance(keys,str):
                js_dat["Argument"]['NODE_ELEMS'] = {"STRUCTURE_GROUP_NAME": keys}


            if loadcase!=[]: js_dat["Argument"]['LOAD_CASE_NAMES'] = loadcase

            if components!=['all']:
                if "Elem" not in components: components.append("Elem")
                if "Load" not in components: components.append("Load")
                if "Part" not in components: components.append("Part")
                if "Component" not in components: components.append("Component")
                js_dat["Argument"]['COMPONENTS'] = components
            
            if items!=['all']:
                js_dat["Argument"]['ITEM_TO_DISPLAY'] = items



            currUNIT = _getUNIT()
            Model.units(force=force_unit,length=len_unit)
            ss_json = MidasAPI("POST","/post/table",js_dat)
            _setUNIT(currUNIT)
            return _JSToDF_ResTable(ss_json)




from ._mapi import MidasAPI



def _ThikADD(self):
    # Commom HERE ---------------------------------------------
    id = int(self.ID)
    if Thickness.ids == []: 
        count = 1
    else:
        count = max(Thickness.ids)+1

    if id==0 :
        self.ID = count
        Thickness.thick.append(self)
        Thickness.ids.append(int(self.ID))
    elif id in Thickness.ids:
        self.ID=int(id)
        print(f'⚠️  Thickness with ID {id} already exist! It will be replaced.')
        index=Thickness.ids.index(id)
        Thickness.thick[index]=self
    else:
        self.ID=id        
        Thickness.thick.append(self)
        Thickness.ids.append(int(self.ID))
    # Common END -------------------------------------------------------


def _updateThik(self):
    js2s = {'Assign':{self.ID : _Obj2JS(self)}}
    MidasAPI('PUT','/db/THIK',js2s)
    return js2s


def _Obj2JS(obj):

    js={
            "NAME": obj.NAME,
            "TYPE": obj.TYPE,
            "bINOUT": obj.bINOUT,
            "T_IN": obj.T_IN,
            "T_OUT": obj.T_OUT,
            "OFFSET": obj.OFF_TYPE,
            "O_VALUE": obj.OFFSET
        }
    return js



def _JS2Obj(id,js):
    name = js['NAME']
    type = js['TYPE']
    binout = js['bINOUT']
    t_in = js['T_IN']
    t_out = js['T_OUT']
    try : offset = js['OFFSET'] 
    except: offset = 0
    off_value = js['O_VALUE']

    t_out2=-1
    if binout:t_out2 = t_out

    if type == 'VALUE':
        Thickness(t_in,t_out2,off_value,offset,name,id)
    


class _common:
    def __str__(self):
        return str(f'ID = {self.ID}  \nJSON : {_Obj2JS(self)}\n')

    def update(self):
        return _updateThik(self)


class Thickness(_common):
    """Create Thicknes"""
    thick = []
    ids = []

    def __init__(self,thick=0.0,thick_out=-1,offset=0,off_type='rat',name="",id=None):
            if id == None: id = 0  
            self.ID = id
            if name == "":
                self.NAME = str(thick)
            else: self.NAME = name
            self.TYPE = 'VALUE'
            self.T_IN = thick
            self.bINOUT = True

            if thick_out==-1:
                self.T_OUT = thick
                self.bINOUT = False
            else: self.T_OUT = thick_out

            self.OFFSET = offset

            if off_type=='rat':
                self.OFF_TYPE = 1
            else: self.OFF_TYPE = 2

            if offset==0:
                self.OFF_TYPE =0

            _ThikADD(self)


    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for sect in cls.thick:
            js = _Obj2JS(sect)
            json["Assign"][sect.ID] = js
        return json
    
    @staticmethod
    def create():
        MidasAPI("PUT","/db/THIK",Thickness.json())
        
    @staticmethod
    def get():
        return MidasAPI("GET","/db/THIK")
    
    
    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/THIK")
        Thickness.thick=[]
        Thickness.ids=[]

    @staticmethod
    def clear():
        Thickness.thick=[]
        Thickness.ids=[]


    @staticmethod
    def sync():
        a = Thickness.get()
        if a != {'message': ''}:
            if list(a['THIK'].keys()) != []:
                Thickness.thick = []
                Thickness.ids=[]
                for sect_id in a['THIK'].keys():
                    _JS2Obj(sect_id,a['THIK'][sect_id])





from ._mapi import MidasAPI,NX
from colorama import Fore,Style
from ._node import Node , NodeLocalAxis
from ._element import Element
from ._group import Group
from ._load import Load
from ._boundary import Boundary

from ._section import Section
from ._material import Material
from ._thickness import Thickness

from ._tendon import Tendon
from ._loadcomb import LoadCombination
from ._movingload import MovingLoad

from ._temperature import Temperature
from ._construction import CS

from collections import defaultdict
class Model:

    #4 Function to check analysis status & perform analysis if not analyzed
    @staticmethod
    def analyse():
        """Checkes whether a model is analyzed or not and then performs analysis if required."""
        json_body = {
        "Argument": {
            "HEIGHT" : 1,
            "WIDTH" : 1,
            "SET_MODE": "post"
        }
        }
        resp = MidasAPI('POST','/view/CAPTURE',json_body)

        if 'message' in resp or 'error' in resp:
                MidasAPI("POST","/doc/ANAL",{"Assign":{}})
                return True
        print(" ⚠️   Model ananlysed. Switching to post-processing mode.")

    #9 Function to remove duplicate nodes and elements from Node & Element classes\
    @staticmethod
    def merge_nodes(tolerance = 0):
        """This functions removes duplicate nodes defined in the Node Class and modifies Element class accordingly.  \nSample: remove_duplicate()"""
        a=[]
        b=[]
        node_json = Node.json()
        elem_json = Element.json()
        node_di = node_json["Assign"]
        elem_di = elem_json["Assign"]
        for i in list(node_di.keys()):
            for j in list(node_di.keys()):
                if list(node_di.keys()).index(j) > list(node_di.keys()).index(i):
                    if (node_di[i]["X"] >= node_di[j]["X"] - tolerance and node_di[i]["X"] <= node_di[j]["X"] + tolerance):
                        if (node_di[i]["Y"] >= node_di[j]["Y"] - tolerance and node_di[i]["Y"] <= node_di[j]["Y"] + tolerance):
                            if (node_di[i]["Z"] >= node_di[j]["Z"] - tolerance and node_di[i]["Z"] <= node_di[j]["Z"] + tolerance):
                                a.append(i)
                                b.append(j)
        for i in range(len(a)):
            for j in range(len(b)):
                if a[i] == b[j]: 
                    a[i] = a[j]
                    for k in elem_di.keys():
                        for i in range(len(a)):
                            if elem_di[k]['NODE'][0] == b[i]: elem_di[k]['NODE'][0] = a[i]
                            if elem_di[k]['NODE'][1] == b[i]: elem_di[k]['NODE'][1] = a[i]
                            try: 
                                if elem_di[k]['NODE'][3] == b[i]: elem_di[k]['NODE'][3] = a[i]
                            except: pass
                            try: 
                                if elem_di[k]['NODE'][4] == b[i]: elem_di[k]['NODE'][4] = a[i]
                            except: pass

        if len(b)>0:
            for i in range(len(b)):
                if b[i] in node_di: del node_di[b[i]]
            Node.nodes = []
            Node.ids = []
            for i in node_di.keys():
                Node(node_di[i]['X'], node_di[i]['Y'], node_di[i]['Z'], i)
            Element.elements = []
            Element.ids = []
            for i in elem_di.keys():
                Element(elem_di[i], i)

    @staticmethod
    def units(force = "KN",length = "M", heat = "BTU", temp = "C"):
        """force --> KN, N, KFG, TONF, LFB, KIPS ||  
        \ndist --> M, CM, MM, FT, IN ||  
        \nheat --> CAL, KCAL, J, KJ, BTU ||  
        \ntemp --> C, F
        \nDefault --> KN, M, BTU, C"""
        if temp not in ["C","F"]:
            temp="C"
        if force not in ["KN", "N", "KGF", "TONF", "LBF", "KIPS"]:
            force = "KN"
        if length not in ["M", "CM", "MM", "FT", "IN"]:
            dist = "M"
        if heat not in ["CAL", "KCAL", "J", "KJ", "BTU"]:
            heat = "BTU"
        unit={"Assign":{
            1:{
                "FORCE":force,
                "DIST":length,
                "HEAT":heat,
                "TEMPER":temp
            }
        }}
        MidasAPI("PUT","/db/UNIT",unit)

    @staticmethod
    def select(crit_1 = "X", crit_2 = 0, crit_3 = 0, st = 'a', en = 'a', tolerance = 0):
        """Get list of nodes/elements as required.\n
        crit_1 (=> Along: "X", "Y", "Z". OR, IN: "XY", "YZ", "ZX". OR "USM"),\n
        crit_2 (=> With Ordinate value: Y value, X value, X Value, Z value, X value, Y value. OR Material ID),\n
        crit_3 (=> At Ordinate 2 value: Z value, Z value, Y value, 0, 0, 0. OR Section ID),\n
        starting ordinate, end ordinate, tolerance, node dictionary, element dictionary.\n
        Sample:  get_select("Y", 0, 2) for selecting all nodes and elements parallel Y axis with X ordinate as 0 and Z ordinate as 2."""
        output = {'NODE':[], 'ELEM':[]}
        ok = 0
        no = Node.json()
        el = Element.json()
        if crit_1 == "USM":
            materials = Material.json()
            sections = Section.json()
            elements = el
            k = list(elements.keys())[0]
            mat_nos = list((materials["Assign"].keys()))
            sect_nos = list((sections["Assign"].keys()))
            elem = {}
            for m in mat_nos:
                elem[int(m)] = {}
                for s in sect_nos:
                        elem[int(m)][int(s)] = []
            for e in elements[k].keys(): elem[((elements[k][e]['MATL']))][((elements[k][e]['SECT']))].append(int(e))
            output['ELEM'] = elem[crit_2][crit_3]
            ok = 1
        elif no != "" and el != "":
            n_key = list(no.keys())[0]
            e_key = list(el.keys())[0]
            if n_key == "Assign": no["Assign"] = {str(key):value for key,value in no["Assign"].items()}
            if e_key == "Assign": el["Assign"] = {str(key):value for key,value in el["Assign"].items()}
            if crit_1 == "X": 
                cr2 = "Y"
                cr3 = "Z"
                ok = 1
            if crit_1 == "Y": 
                cr2 = "X"
                cr3 = "Z"
                ok = 1
            if crit_1 == "Z": 
                cr2 = "X"
                cr3 = "Y"
                ok = 1
            if crit_1 == "XY" or crit_1 == "YX":
                cr2 = "Z"
                ok = 1
            if crit_1 == "YZ" or crit_1 == "ZY":
                cr2 = "X"
                ok = 1
            if crit_1 == "ZX" or crit_1 == "XZ":
                cr2 = "Y"
                ok = 1
            if len(crit_1) == 1 and ok == 1:
                if st == 'a': st = min([v[crit_1] for v in no[n_key].values()])
                if en == 'a': en = max([v[crit_1] for v in no[n_key].values()])
                for n in no[n_key].keys():
                    curr = no[n_key][n]
                    if curr[cr2] >= crit_2 - tolerance and curr[cr2] <= crit_2 + tolerance:
                        if curr[cr3] >= crit_3 - tolerance and curr[cr3] <= crit_3 + tolerance:
                            if curr[crit_1] >= st and curr[crit_1] <= en: output['NODE'].append(int(n))
                for e in el[e_key].keys():
                    curr_0 = no[n_key][str(el[e_key][e]['NODE'][0])]
                    curr_1 = no[n_key][str(el[e_key][e]['NODE'][1])]
                    if curr_0[cr2] == curr_1[cr2] and curr_0[cr3] == curr_1[cr3]:
                        if curr_0[cr2] >= crit_2 - tolerance and curr_0[cr2] <= crit_2 + tolerance:
                            if curr_0[cr3] >= crit_3 - tolerance and curr_0[cr3] <= crit_3 + tolerance:
                                if curr_1[cr2] >= crit_2 - tolerance and curr_1[cr2] <= crit_2 + tolerance:
                                    if curr_1[cr3] >= crit_3 - tolerance and curr_1[cr3] <= crit_3 + tolerance:
                                        if curr_0[crit_1] >= st and curr_0[crit_1] <= en and curr_1[crit_1] >= st and curr_1[crit_1] <= en:
                                            output['ELEM'].append(int(e))
            if len(crit_1) == 2 and ok == 1:
                if st == 'a': st = min(min([v[crit_1[0]] for v in no[n_key].values()]), min([v[crit_1[1]] for v in no[n_key].values()]))
                if en == 'a': en = max(max([v[crit_1[0]] for v in no[n_key].values()]), max([v[crit_1[1]] for v in no[n_key].values()]))
                for n in no[n_key].keys():
                    curr = no[n_key][n]
                    if curr[cr2] >= crit_2 - tolerance and curr[cr2] <= crit_2 + tolerance:
                        if curr[crit_1[0]] >= st and curr[crit_1[1]] >= st and curr[crit_1[0]] <= en and curr[crit_1[1]] <= en: output['NODE'].append(int(n))
                for e in el[e_key].keys():
                    curr_0 = no[n_key][str(el[e_key][e]['NODE'][0])]
                    curr_1 = no[n_key][str(el[e_key][e]['NODE'][1])]
                    if curr_0[cr2] == curr_1[cr2]:
                        if curr_0[cr2] >= crit_2 - tolerance and curr_0[cr2] <= crit_2 + tolerance:
                            if curr_1[cr2] >= crit_2 - tolerance and curr_1[cr2] <= crit_2 + tolerance:
                                if curr_0[crit_1[0]] >= st and curr_0[crit_1[0]] <= en and curr_1[crit_1[0]] >= st and curr_1[crit_1[0]] <= en:
                                    if curr_0[crit_1[1]] >= st and curr_0[crit_1[1]] <= en and curr_1[crit_1[1]] >= st and curr_1[crit_1[1]] <= en:
                                        output['ELEM'].append(int(e))
        if ok != 1: output = "Incorrect input.  Please check the syntax!"
        return output



    # @staticmethod
    # def _create2(request = "update", set = 1, force = "KN", length = "M", heat = "BTU", temp = "C"):
    #     """request["update" to update a model, "call" to get details of existing model], \nforce[Optional], length[Optional], heat[Optional], temp[Optional].  
    #     \nSample: model() to update/create model. model("call") to get details of existing model and update classes.\n
    #     set = 1 => Functions that don't need to call data from connected model file.\n
    #     set = 2 => Functions that may need to call data from connected model file."""
    #     Model.units(force, length, heat, temp)
    #     if MAPI_KEY.data == []:  print(f"Enter the MAPI key using the MAPI_KEY command.")
    #     if MAPI_KEY.data != []:
    #         if set == 1:
    #             if request == "update" or request == "create" or request == "PUT":
    #                 if Node.json() != {"Assign":{}}: Node.create()
    #                 if Element.json() != {"Assign":{}}: Element.create()
    #                 if Section.json() != {"Assign":{}}: Section.create()
    #                 if Group.json_BG() != {"Assign":{}}: Group.create_BG()
    #                 if Group.json_LG() != {"Assign":{}}: Group.create_LG()
    #                 if Group.json_TG() != {"Assign":{}}: Group.create_TG()
    #                 if Material.json() != {"Assign":{}}: Material.create()
    #             if request == "call" or request == "GET":
    #                 Node.sync()
    #                 Element.sync()
    #                 Section.sync()
    #                 Group.sync()
    #                 Material.sync()
    #         if set == 2:
    #             if request == "update" or request == "create" or request == "PUT":
    #                 if Node.json() != {"Assign":{}}: Node.create()
    #                 if Element.json() != {"Assign":{}}: Element.create()
    #                 if Section.json() != {"Assign":{}}: Section.create()
    #                 if Group.json_BG() != {"Assign":{}}: Group.create_BG()
    #                 if Group.json_LG() != {"Assign":{}}: Group.create_LG()
    #                 if Group.json_TG() != {"Assign":{}}: Group.create_TG()
    #                 if Material.json() != {"Assign":{}}: Material.create()
    #                 if Group.json_SG() != {"Assign":{}}: Group.create_SG()
    #             if request == "call" or request == "GET": 
    #                 Node.update_class()
    #                 Element.update_class()
    #                 Section.update_class()
    #                 Group.update_class()
    #                 Material.update_class()


    @staticmethod
    def maxID(dbNAME:str = 'NODE') -> int :
        ''' 
        Returns maximum ID of a DB in CIVIL NX
        dbNAME - 'NODE' , 'ELEM' , 'THIK' , 'SECT' 
        If no data exist, 0 is returned
        '''
        dbJS = MidasAPI('GET',f'/db/{dbNAME}')
        if dbJS == {'message': ''}:
            return 0
        return max(map(int, list(dbJS[dbNAME].keys())))

    @staticmethod
    def create():
        """Create Material, Section, Node, Elements, Groups and Boundary."""
        from tqdm import tqdm
        pbar = tqdm(total=15,desc="Creating Model...")

        if Material.mats!=[]: Material.create()
        pbar.update(1)
        pbar.set_description_str("Creating Section...")
        if Section.sect!=[]: Section.create()
        pbar.update(1)
        pbar.set_description_str("Creating Thickness...")
        if Thickness.thick!=[]: Thickness.create()
        pbar.update(1)
        pbar.set_description_str("Creating Node...")
        if Node.nodes!=[]: Node.create()
        pbar.update(1)
        pbar.set_description_str("Creating Element...")
        if Element.elements!=[] : Element.create()
        pbar.update(1)
        pbar.set_description_str("Creating Node Local Axis...")
        if NodeLocalAxis.skew!=[] : NodeLocalAxis.create()
        pbar.update(1)
        pbar.set_description_str("Creating Group...")
        Group.create()
        pbar.update(1)
        pbar.set_description_str("Creating Boundary...")
        Boundary.create()
        pbar.update(1)
        pbar.set_description_str("Creating Load...")
        Load.create()
        pbar.update(1)
        pbar.set_description_str("Creating Temperature...")
        Temperature.create()
        pbar.update(1)
        pbar.set_description_str("Creating Tendon...")
        Tendon.create()
        pbar.update(1)
        pbar.set_description_str("Creating Tapered Group...")
        if Section.TaperedGroup.data !=[] : Section.TaperedGroup.create()
        pbar.update(1)
        pbar.set_description_str("Creating Construction Stages...")
        CS.create()
        pbar.update(1)
        pbar.set_description_str("Creating Moving Load...")
        MovingLoad.create()
        pbar.update(1)
        pbar.set_description_str("Creating Load Combination...")
        LoadCombination.create()
        pbar.update(1)
        pbar.set_description_str(Fore.GREEN+"Model creation complete"+Style.RESET_ALL)
        





    @staticmethod
    def clear():
        Material.clearAll()
        Section.clear()
        Thickness.clear()
        Node.clear()
        Element.clear()
        NodeLocalAxis.clear()
        Group.clear()
        Boundary.clear()
        Load.clear()
        Temperature.clear()
        Tendon.clear()
        Section.TaperedGroup.clear()
        LoadCombination.clear()



    @staticmethod
    def type(strc_type=0,mass_type=1,gravity:float=0,mass_dir=1):
        """Structure Type option 
        --------------------------------
        
        Structure Type:
            0 = 3D
            1 = X-Z Plane
            2 = Y-Z Plane
            3 = X-Y Plane
            4 = Constraint RZ

        Mass Type:
            1 = Lumped Mass
            2 = Consistent Mass
        
        Gravity Acceleration (g) = 9.81 m/s^2
        
        Mass Direction(Structure Mass type):
            1 = Convert to X, Y, Z
            2 = Convert to X, Y
            3 = Convert to Z
        """

        js = {"Assign": {
              "1":{}}}
        

        js["Assign"]["1"]["STYP"] = strc_type

        js["Assign"]["1"]["MASS"] = mass_type

        if mass_dir==0:
            js["Assign"]["1"]["bSELFWEIGHT"] = False
        else:
            js["Assign"]["1"]["bSELFWEIGHT"] = True
            js["Assign"]["1"]["SMASS"] = mass_dir

        if gravity!=0:
            js["Assign"]["1"]["GRAV"] = gravity


        MidasAPI("PUT","/db/STYP",js)

    @staticmethod
    def save(location=""):
        """Saves the model\nFor the first save, provide location - \nModel.save("D:\\model2.mcb")"""
        if location=="":
            MidasAPI("POST","/doc/SAVE",{"Argument":{}})
        else:
            if location.endswith('.mcb') or location.endswith('.mcbz'):
                MidasAPI("POST","/doc/SAVEAS",{"Argument":str(location)})#Dumy location
            else:
                print('⚠️  File extension is missing')
                

    @staticmethod
    def saveAs(location=""):
        """Saves the model at location provided   
         Model.saveAs("D:\\model2.mcb")"""
        if location.endswith('.mcb') or location.endswith('.mcbz'):
            MidasAPI("POST","/doc/SAVEAS",{"Argument":str(location)})
        else:
            print('⚠️  File extension is missing')
    
    @staticmethod
    def open(location=""):
        """Open Civil NX model file \n Model.open("D:\\model.mcb")"""
        if location.endswith('.mcb') or location.endswith('.mcbz'):
            MidasAPI("POST","/doc/OPEN",{"Argument":str(location)})
        else:
            print('⚠️  File extension is missing')
        

    @staticmethod
    def new():
        """Creates a new model"""
        MidasAPI("POST","/doc/NEW",{"Argument":{}})

    @staticmethod
    def info(project_name="",revision="",user="",title=""):
        """Enter Project information"""

        js = {"Assign": {
              "1":{}}}
        
        if project_name+revision+user+title=="":
            return MidasAPI("GET","/db/PJCF",{})
        else:
            if project_name!="":
                js["Assign"]["1"]["PROJECT"] = project_name
            if revision!="":
                js["Assign"]["1"]["REVISION"] = revision
            if user!="":
                js["Assign"]["1"]["USER"] = user
            if title!="":
                js["Assign"]["1"]["TITLE"] = title


            MidasAPI("PUT","/db/PJCF",js)
    
    @staticmethod
    def exportJSON(location=""):
        """Export the model data as JSON file
        Model.exportJSON('D:\\model.json')"""
        if location.endswith('.json'):
            MidasAPI("POST","/doc/EXPORT",{"Argument":str(location)})
        else:
            print('⚠️  Location data in exportJSON is missing file extension')

    @staticmethod
    def exportMCT(location=""):
        """Export the model data as MCT file
        Model.exportMCT('D:\\model.mct')"""
        if location.endswith('.mct'):
            MidasAPI("POST","/doc/EXPORTMXT",{"Argument":str(location)})
        else:
            print('⚠️  Location data in exportMCT is missing file extension')

    
    @staticmethod
    def importJSON(location=""):
        """Import JSON data file in MIDAS CIVIL NX
        Model.importJSON('D:\\model.json')"""
        if location.endswith('.json'):
            MidasAPI("POST","/doc/IMPORT",{"Argument":str(location)})
        else:
            print('⚠️  Location data in importJSON is missing file extension')

    @staticmethod
    def importMCT(location=""):
        """Import MCT data file in MIDAS CIVIL NX
        Model.importMCT('D:\\model.mct')"""
        if location.endswith('.mct'):
            MidasAPI("POST","/doc/IMPORTMXT",{"Argument":str(location)})
        else:
            print('⚠️  Location data in importMCT is missing file extension')

    @staticmethod
    def get_element_connectivity():
        element_connectivity = {}
        for element in Element.elements:
            element_id = element.ID
            connected_nodes = element.NODE
            element_connectivity.update({element_id: connected_nodes})
        return element_connectivity

    @staticmethod
    def get_node_connectivity():
        element_connectivity = Model.get_element_connectivity()
        node_connectivity = defaultdict(list)

        for element_id, nodes in element_connectivity.items():
            for node in nodes:
                node_connectivity[node].append(element_id)
        node_connectivity = dict(node_connectivity)
        return node_connectivity

    @staticmethod
    def visualise():
        if NX.visualiser:
            try:
                from ._visualise import displayWindow
                displayWindow()
            except:
                pass

    @staticmethod
    def snap():
        if NX.visualiser:
            try:
                from ._visualise import take_snapshot
                take_snapshot()
            except:
                pass











import requests
from colorama import Fore,Style
from ._mapi import NX,MidasAPI,MAPI_KEY,MAPI_BASEURL,MAPI_COUNTRY,Midas_help
_version_ = "1.2.9"


print('\n╭────────────────────────────────────────────────────────────────────────────────────╮')
print(Style.BRIGHT+f'│                      MIDAS CIVIL-NX PYTHON LIBRARY v{_version_}  🐍                      │')
print('╰────────────────────────────────────────────────────────────────────────────────────╯\n'+Style.RESET_ALL)

if NX.version_check:
    resp =  requests.get("https://pypi.org/pypi/midas_civil/json").json()
    latest_ver =  resp["info"]["version"]
    if _version_ != latest_ver:        
        print(Fore.YELLOW +'╭─ ⚠️   ──────────────────────────────────────────────────────────────────────────────╮')
        print(f"│    Warning: You are using v{_version_}, but the latest available version is v{latest_ver}.      │")
        print(f"│    Run 'pip install midas_civil --upgrade' to update.                              │")
        print('╰────────────────────────────────────────────────────────────────────────────────────╯\n'+Style.RESET_ALL)


from ._model import Model
from ._boundary import Boundary
from ._utils import getID,getNodeID,utils
from ._node import Node,nodeByID,closestNode,NodeLocalAxis,nodesInRadius
from ._element import Element,elemByID  #Revise
from ._load import Load,Load_Case   # Revise it
from ._group import nodesInGroup,elemsInGroup,Group
from ._loadcomb import LoadCombination


from ._material import Material,CompStrength,CreepShrinkage,TDMatLink

from ._section import Section,Offset

from ._construction import CS
from ._thickness import Thickness
from ._temperature import Temperature # dueto LoadCase

from ._tendon import Tendon
from ._view import View,ResultGraphic,Image

from ._movingload import MovingLoad
from ._settlement import Settlement
from ._analysiscontrol import AnalysisControl
from ._BoundaryChangeAssignment import BoundaryChangeAssignment         # <=== NEEDS A REVIEW (UNNECESSARY CALL)

from ._result_test import Result  #Too much time taken here




from ._offsetSS import Offset
from ._offsetSS import _common
import math



class _SS_PSC_12CELL(_common):
    def __init__(self,Name='',Shape='1CEL',Joint=[0,0,0,0,0,0,0,0],
                    HO1=0,HO2=0,HO21=0,HO22=0,HO3=0,HO31=0,
                    BO1=0,BO11=0,BO12=0,BO2=0,BO21=0,BO3=0,
                    HI1=0,HI2=0,HI21=0,HI22=0,HI3=0,HI31=0,HI4=0,HI41=0,HI42=0,HI5=0,
                    BI1=0,BI11=0,BI12=0,BI21=0,BI3=0,BI31=0,BI32=0,BI4=0,
                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=0):
                
        self.ID = id
        self.NAME = Name
        self.SHAPE = Shape
        self.TYPE = 'PSC'

        self.JO1=bool(Joint[0])
        self.JO2=bool(Joint[1])
        self.JO3=bool(Joint[2])
        self.JI1=bool(Joint[3])
        self.JI2=bool(Joint[4])
        self.JI3=bool(Joint[5])
        self.JI4=bool(Joint[6])
        self.JI5=bool(Joint[7])

        self.OFFSET = Offset
        self.USESHEAR = bool(useShear)
        self.USE7DOF = bool(use7Dof)

        self.HO1 = HO1
        self.HO2 = HO2
        self.HO21 = HO21
        self.HO22= HO22
        self.HO3 = HO3
        self.HO31 = HO31

        self.BO1 = BO1
        self.BO11 = BO11
        self.BO12 = BO12
        self.BO2 = BO2
        self.BO21 = BO21
        self.BO3 = BO3

        self.HI1 = HI1
        self.HI2 = HI2
        self.HI21 = HI21
        self.HI22 = HI22
        self.HI3 = HI3
        self.HI31 = HI31
        self.HI4 = HI4
        self.HI41 = HI41
        self.HI42 = HI42
        self.HI5 = HI5

        self.BI1 = BI1
        self.BI11 = BI11
        self.BI12 = BI12
        self.BI21 = BI21
        self.BI3 = BI3
        self.BI31 = BI31
        self.BI32 = BI32
        self.BI4 = BI4

    
    def __str__(self):
         return f'  >  ID = {self.ID}   |  PSC 1-2 CELL SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                "SECTTYPE": "PSC",
                "SECT_NAME": sect.NAME,
                "SECT_BEFORE": {
                    "SHAPE": sect.SHAPE,
                    "SECT_I": {
                        "vSIZE_PSC_A": [sect.HO1,sect.HO2,sect.HO21,sect.HO22,sect.HO3,sect.HO31],
                        "vSIZE_PSC_B": [sect.BO1,sect.BO11,sect.BO12,sect.BO2,sect.BO21,sect.BO3,],
                        "vSIZE_PSC_C": [sect.HI1,sect.HI2,sect.HI21,sect.HI22,sect.HI3,sect.HI31,sect.HI4,sect.HI41,sect.HI42,sect.HI5],
                        "vSIZE_PSC_D": [sect.BI1,sect.BI11,sect.BI12,sect.BI21,sect.BI3,sect.BI31,sect.BI32,sect.BI4]
                    },
                    "WARPING_CHK_AUTO_I": True,
                    "WARPING_CHK_AUTO_J": True,
                    "SHEAR_CHK": True,
                    "WARPING_CHK_POS_I": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "WARPING_CHK_POS_J": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "USE_AUTO_SHEAR_CHK_POS": [[True,False,True],[False,False,False]],
                    "USE_WEB_THICK_SHEAR": [[True, True,True],[False,False,False]],
                    "SHEAR_CHK_POS": [[0,0,0],[0,0,0]],
                    "USE_WEB_THICK": [True,False],
                    "WEB_THICK": [0,0],
                    "JOINT": [sect.JO1,sect.JO2,sect.JO3,sect.JI1,sect.JI2,sect.JI3,sect.JI4,sect.JI5]
                }
            }
        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    

    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):
        #--- PSC 1,2 CELL -------------------
        vA = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_A']
        vB = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_B']
        vC = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_C']
        vD = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_D']
        joint = js['SECT_BEFORE']['JOINT']
        return _SS_PSC_12CELL(name,shape,joint,
                            vA[0],vA[1],vA[2],vA[3],vA[4],vA[5],
                            vB[0],vB[1],vB[2],vB[3],vB[4],vB[5],
                            vC[0],vC[1],vC[2],vC[3],vC[4],vC[5],vC[6],vC[7],vC[8],vC[9],
                            vD[0],vD[1],vD[2],vD[3],vD[4],vD[5],vD[6],vD[7],
                            offset,uShear,u7DOF,id)
    
    def _centerLine(shape,*args):
        if shape.SHAPE == '1CEL':
            BO1 = shape.BO1
            BO2 = shape.BO2
            BO3 = shape.BO3
            HO1 = shape.HO1
            HO2 = shape.HO2
            HO3 = shape.HO3
            HI1 = shape.HI1
            HI5 = shape.HI5
            BI3 = shape.BI3

            sect_lin_con = [[1,2],[2,3],[3,4],[2,5],[3,6],[5,6]]

            sect_cg_LT = [0,0]
            sect_cg_RB = [2*(BO1+BO2+BO3),-(HO1+HO2+HO3)]
            sect_cg_CC = [(BO1+BO2+BO3),-(HO1+HO2+HO3)/2]

            sect_shape = [[0,0],[BO1,0],[BO1+2*(BO2+BO3),0],[2*(BO1+BO2+BO3),0],[BO1+BO2,-HO1-HO3-HO2],[BO1+BO2+2*BO3,-HO1-HO3-HO2]]
            sect_thk = [HO1,HI1,HO1,BO3-BI3,BO3-BI3,HI5]
            sect_thk_off = [-HO1/2,-HI1/2,-HO1/2,(BO3-BI3)/2,-(BO3-BI3)/2,HI5/2]       



        sect_cg = (sect_cg_LT,sect_cg_CC,sect_cg_RB)

        return sect_shape, sect_thk ,sect_thk_off, sect_cg , sect_lin_con

    








class _SS_PSC_I(_common):
    def __init__(self,Name='',Symm = True,Joint=[0,0,0,0,0,0,0,0,0],
                            H1=0,
                            HL1=0,HL2=0,HL21=0,HL22=0,HL3=0,HL4=0,HL41=0,HL42=0,HL5=0,
                            BL1=0,BL2=0,BL21=0,BL22=0,BL4=0,BL41=0,BL42=0,

                            HR1=0,HR2=0,HR21=0,HR22=0,HR3=0,HR4=0,HR41=0,HR42=0,HR5=0,
                            BR1=0,BR2=0,BR21=0,BR22=0,BR4=0,BR41=0,BR42=0,

                            Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=0):
                
        self.ID = id
        self.NAME = Name
        self.SHAPE = 'PSCI'
        self.TYPE = 'PSC'

        self.SYMM = bool(Symm)

        self.J1=bool(Joint[0])
        self.JL1=bool(Joint[1])
        self.JL2=bool(Joint[2])
        self.JL3=bool(Joint[3])
        self.JL4=bool(Joint[4])

        if self.SYMM:
            self.JR1=bool(Joint[1])
            self.JR2=bool(Joint[2])
            self.JR3=bool(Joint[3])
            self.JR4=bool(Joint[4])

            self.HR1	  =	HL1
            self.HR2	  =	HL2
            self.HR21	  =	HL21
            self.HR22	  =	HL22
            self.HR3	  =	HL3
            self.HR4	  =	HL4
            self.HR41	  =	HL41
            self.HR42	  =	HL42
            self.HR5	  =	HL5

            self.BR1	  =	BL1
            self.BR2	  =	BL2
            self.BR21	  =	BL21
            self.BR22	  =	BL22
            self.BR4	  =	BL4
            self.BR41	  =	BL41
            self.BR42	  =	BL42
        else:
            self.JR1=bool(Joint[5])
            self.JR2=bool(Joint[6])
            self.JR3=bool(Joint[7])
            self.JR4=bool(Joint[8])

            self.HR1	  =	HR1
            self.HR2	  =	HR2
            self.HR21	  =	HR21
            self.HR22	  =	HR22
            self.HR3	  =	HR3
            self.HR4	  =	HR4
            self.HR41	  =	HR41
            self.HR42	  =	HR42
            self.HR5	  =	HR5

            self.BR1	  =	BR1
            self.BR2	  =	BR2
            self.BR21	  =	BR21
            self.BR22	  =	BR22
            self.BR4	  =	BR4
            self.BR41	  =	BR41
            self.BR42	  =	BR42

        self.OFFSET = Offset
        self.USESHEAR = bool(useShear)
        self.USE7DOF = bool(use7Dof)

        self.H1	  =	H1
        self.HL1	  =	HL1
        self.HL2	  =	HL2
        self.HL21	  =	HL21
        self.HL22	  =	HL22
        self.HL3	  =	HL3
        self.HL4	  =	HL4
        self.HL41	  =	HL41
        self.HL42	  =	HL42
        self.HL5	  =	HL5

        self.BL1	  =	BL1
        self.BL2	  =	BL2
        self.BL21	  =	BL21
        self.BL22	  =	BL22
        self.BL4	  =	BL4
        self.BL41	  =	BL41
        self.BL42	  =	BL42 

    
    def __str__(self):
         return f'  >  ID = {self.ID}   |  PSC I SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                "SECTTYPE": "PSC",
                "SECT_NAME": sect.NAME,
                "SECT_BEFORE": {
                    "SHAPE": sect.SHAPE,
                    "SECT_I": {
                        "vSIZE_PSC_A": [sect.H1,sect.HL1,sect.HL2,sect.HL21,sect.HL22,sect.HL3,sect.HL4,sect.HL41,sect.HL42,sect.HL5],
                        "vSIZE_PSC_B": [sect.BL1,sect.BL2,sect.BL21,sect.BL22,sect.BL4,sect.BL41,sect.BL42],
                        "vSIZE_PSC_C": [sect.HR1,sect.HR2,sect.HR21,sect.HR22,sect.HR3,sect.HR4,sect.HR41,sect.HR42,sect.HR5],
                        "vSIZE_PSC_D": [sect.BR1,sect.BR2,sect.BR21,sect.BR22,sect.BR4,sect.BR41,sect.BR42]
                    },
                    "WARPING_CHK_AUTO_I": True,
                    "WARPING_CHK_AUTO_J": True,
                    "SHEAR_CHK": True,
                    "WARPING_CHK_POS_I": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "WARPING_CHK_POS_J": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "USE_AUTO_SHEAR_CHK_POS": [[True,False,True],[False,False,False]],
                    "USE_WEB_THICK_SHEAR": [[True, True,True],[False,False,False]],
                    "SHEAR_CHK_POS": [[0,0,0],[0,0,0]],
                    "USE_WEB_THICK": [True,False],
                    "WEB_THICK": [0,0],
                    "USE_SYMMETRIC" : sect.SYMM,
                    "JOINT": [sect.J1,sect.JL1,sect.JL2,sect.JL3,sect.JL4,sect.JR1,sect.JR2,sect.JR3,sect.JR4]
                }
            }
        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    

    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):
        symm = js['SECT_BEFORE']['USE_SYMMETRIC']
        vA = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_A']
        vB = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_B']
        vC = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_C']
        vD = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_D']
        joint = js['SECT_BEFORE']['JOINT']
        return _SS_PSC_I(name,symm,joint,
                            vA[0],
                            vA[1],vA[2],vA[3],vA[4],vA[5],vA[6],vA[7],vA[8],vA[9],
                            vB[0],vB[1],vB[2],vB[3],vB[4],vB[5],vB[6],
                            vC[0],vC[1],vC[2],vC[3],vC[4],vC[5],vC[6],vC[7],vC[8],
                            vD[0],vD[1],vD[2],vD[3],vD[4],vD[5],vD[6],
                            offset,uShear,u7DOF,id)
    











import numpy as np
def _poly_dir(poly,rot='CCW'):
    outer_cg = np.mean(poly,axis=0)
    outer_t = np.subtract(poly,outer_cg)
    dir = 0
    for i in range(len(poly)-1):
        dir+=outer_t[i][0]*outer_t[i+1][1]-outer_t[i][1]*outer_t[i+1][0]
    if dir < 0:
        poly.reverse()
    
    if rot == 'CW':
        poly.reverse()

    return poly






class _SS_PSC_Value(_common):
    def __init__(self,Name:str,
                    OuterPolygon:list,InnerPolygon:list=[],
                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=0):
        
        '''
            Outer Polygon -> List of points ; Last input is different from first
                [(0,0),(1,0),(1,1),(0,1)]
            Inner Polygon -> List of points ; Last input is different from first
                Only one inner polygon
        '''
        
        self.ID = id
        self.NAME = Name
        self.SHAPE = 'VALUE'
        self.TYPE = 'PSC'

        self.OFFSET = Offset
        self.USESHEAR = bool(useShear)
        self.USE7DOF = bool(use7Dof)

        self.OUTER_POLYGON = _poly_dir(OuterPolygon)
        self.INNER_POLYGON = []
        self.N_INNER_POLYGON = 0

        temp_arr = [] 

        # Finding no. of internal polygons
        if InnerPolygon != []:
            if not isinstance(InnerPolygon[0][0],(int,float)):
                self.N_INNER_POLYGON = len(InnerPolygon)
                temp_arr = InnerPolygon 
                
            else:
                temp_arr.append(InnerPolygon) #Convert to list
                self.N_INNER_POLYGON = 1

        for i in range(len(temp_arr)):
            self.INNER_POLYGON.append(_poly_dir(temp_arr[i],'CW'))


    def __str__(self):
         return f'  >  ID = {self.ID}   |  PSC VALUE SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                    "SECTTYPE": "PSC",
                    "SECT_NAME": sect.NAME,
                    "CALC_OPT": True,
                    "SECT_BEFORE": {
                        "SHAPE": "VALU",
                        "SECT_I": {
                            "SECT_NAME": "",
                            "vSIZE": [0.1, 0.1, 0.1, 0.1],
                            "OUTER_POLYGON": [
                                {
                                    "VERTEX": [
                                        {"X": 5, "Y": 5},
                                        {"X": -5, "Y": 5}
                                    ]
                                }
                            ]
                        },
                        "SHEAR_CHK": True,
                        "SHEAR_CHK_POS": [[0.1, 0, 0.1], [0, 0, 0]],
                        "USE_AUTO_QY": [[True, True, True], [False, False, False]],
                        "WEB_THICK": [0, 0],
                        "USE_WEB_THICK_SHEAR": [[True, True, True], [False, False, False]]
                    }
                }
        
        v_list = []
        for i in sect.OUTER_POLYGON:
            v_list.append({"X":i[0],"Y":i[1]})
        js["SECT_BEFORE"]["SECT_I"]["OUTER_POLYGON"][0]["VERTEX"] =v_list

        

        if sect.N_INNER_POLYGON > 0 :

            js["SECT_BEFORE"]["SECT_I"]["INNER_POLYGON"]= []

            mult_ver = []
            for n in range(sect.N_INNER_POLYGON):
                vi_list = []

                js["SECT_BEFORE"]["SECT_I"]["INNER_POLYGON"]= [
                    {
                        "VERTEX": []
                    }
                ]
                for i in sect.INNER_POLYGON[n]:
                    vi_list.append({"X":i[0],"Y":i[1]})

                ver_json = {"VERTEX": vi_list}
                mult_ver.append(ver_json)

            js["SECT_BEFORE"]["SECT_I"]["INNER_POLYGON"] = mult_ver

        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    

    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):

        outer_pt = []
        for pt in js["SECT_BEFORE"]["SECT_I"]["OUTER_POLYGON"][0]["VERTEX"]:
            outer_pt.append((pt['X'],pt['Y']))

        inner_pt = []
        if 'INNER_POLYGON' in js["SECT_BEFORE"]["SECT_I"]:
            innerJSON = js["SECT_BEFORE"]["SECT_I"]['INNER_POLYGON']
            for n_holes in innerJSON:
                h_pt = []
                for pt in n_holes['VERTEX']:
                    h_pt.append([pt['X'],pt['Y']])
                inner_pt.append(h_pt)

        return _SS_PSC_Value(name,outer_pt,inner_pt,offset,uShear,u7DOF,id)


from ._mapi import MidasAPI
from ._load import Load_Case


class BoundaryChangeAssignment:
    
    data = []
    
    def __init__(self,
                 # Support options
                 bSPT: bool = False,
                 bSPR: bool = False, 
                 bGSPR: bool = False,
                 bCGLINK: bool = False,
                 bSSSF: bool = False,
                 bPSSF: bool = False,
                 bRLS: bool = False,
                 bWSSF: bool = False,
                 bESSF: bool = False,
                 bCDOF: bool = False,
                 # Boundary settings
                 vBOUNDARY: list = None,
                 # Load analysis settings for ST type
                 ST_load_assignments: list = None,
                 # Load analysis settings for other types
                 MV: str = None,
                 SM: str = None,
                 THRSEV: str = None,
                 PO: str = None,
                 THNS: str = None,
                 ULAT: str = None):
        """
        Boundary Change Assignment constructor.
        
        Parameters:
            bSPT: Support (default False)
            bSPR: Point Spring Support (default False)
            bGSPR: General Spring Support (default False)
            bCGLINK: Change General Link Property (default False)
            bSSSF: Section Stiffness Scale Factor (default False)
            bPSSF: Plate Stiffness Scale Factor (default False)
            bRLS: Beam End Release (default False)
            bWSSF: Wall Stiffness Scale Factor (default False)
            bESSF: Element Stiffness Scale Factor (default False)
            bCDOF: Constrain DOF associated with specified displacements/Settlements by boundary group combinations (default False)
            vBOUNDARY: List of boundary assignments in format [["L1", "BG2"], ["L2", "BG1"]]
            ST_load_assignments: List of ST type load cases with BGCNAME assignments in format [["Self-weight", "L1"], ["SIDL", "UNCHANGED"]]
            MV, SM, THRSEV, PO, THNS, ULAT: Boundary group names for respective analysis types
            
        Examples:
            # Basic boundary change assignment
            BoundaryChangeAssignment(
                bSPT=True,
                bCDOF=True,
                vBOUNDARY=[["L1", "BG2"], ["L2", "BG1"]],
                ST_load_assignments=[["Self-weight", "L1"]],
                MV="L1"
            )
            
            # Complex assignment with multiple load types
            BoundaryChangeAssignment(
                bSPT=True,
                bSPR=False,
                bGSPR=False,
                vBOUNDARY=[["L1", "BG2"], ["L2", "BG1"]],
                ST_load_assignments=[["Self-weight", "L1"], ["SIDL", "L2"]],
                MV="L1",
                SM="L2",
                THRSEV="L1"
            )
        """
        
        # ID is always 1 based on the pattern
        self.ID = 1
        
        # Set support parameters
        self.bSPT = bSPT
        self.bSPR = bSPR
        self.bGSPR = bGSPR
        self.bCGLINK = bCGLINK
        self.bSSSF = bSSSF
        self.bPSSF = bPSSF
        self.bRLS = bRLS
        self.bWSSF = bWSSF
        self.bESSF = bESSF
        self.bCDOF = bCDOF
        
        # Process boundary data
        self.vBOUNDARY = self._process_boundary_data(vBOUNDARY)
        
        # Process load analysis data
        self.vLOADANAL = self._process_load_analysis_data(
            ST_load_assignments, MV, SM, THRSEV, PO, THNS, ULAT
        )
        
        # Add to static list
        BoundaryChangeAssignment.data.append(self)
        
        # Automatically execute the data when instance is created
        self._execute()
    
    def _process_boundary_data(self, vBOUNDARY):
        """
        Process boundary data from list format to required JSON structure.
        Input: [["L1", "BG2"], ["L2", "BG1"]]
        Output: [{"BGCNAME": "L1", "vBG": ["BG2"]}, {"BGCNAME": "L2", "vBG": ["BG1"]}]
        """
        if not vBOUNDARY:
            return []
        
        boundary_list = []
        for boundary_pair in vBOUNDARY:
            if len(boundary_pair) == 2:
                boundary_list.append({
                    "BGCNAME": boundary_pair[0],
                    "vBG": [boundary_pair[1]]
                })
        
        return boundary_list
    
    def _get_load_cases(self):
        """
        Get load cases from the system using Load_Case.get() command.
        This simulates the API call that would retrieve current load cases.
        """
        try:
            # This would be replaced with actual API call: Load_Case.get()
            # For now, using the provided example data structure
            load_cases_response = Load_Case.get()
            return load_cases_response.get('STLD', {})
        except:
            # Fallback to empty if API call fails
            return {}
    
    def _convert_st_assignments_to_dict(self, ST_load_assignments):
        """
        Convert ST_load_assignments from list format to dictionary format.
        Input: [["Self-weight", "L1"], ["SIDL", "UNCHANGED"]]
        Output: {"Self-weight": "L1", "SIDL": "UNCHANGED"}
        """
        if not ST_load_assignments:
            return {}
        
        assignments_dict = {}
        for assignment_pair in ST_load_assignments:
            if len(assignment_pair) == 2:
                assignments_dict[assignment_pair[0]] = assignment_pair[1]
        
        return assignments_dict
    
    def _process_load_analysis_data(self, ST_load_assignments, MV, SM, THRSEV, PO, THNS, ULAT):
        """
        Process load analysis data combining user input with system load cases.
        """
        load_anal_list = []
        
        # Get current load cases from system
        load_cases = self._get_load_cases()
        
        # Process ST type load cases
        st_cases = {case_data['NAME']: case_data for case_data in load_cases.values() }  # Assuming ST cases are USER type
        
        # Convert ST_load_assignments from list format to dictionary format for internal processing
        st_assignments_dict = self._convert_st_assignments_to_dict(ST_load_assignments)
        
        if st_cases:
            for case_name in st_cases.keys():
                bgcname = "UNCHANGED"  # Default value
                
                # If user provided specific assignment for this load case
                if st_assignments_dict and case_name in st_assignments_dict:
                    bgcname = st_assignments_dict[case_name]
                
                load_anal_list.append({
                    "TYPE": "ST",
                    "BGCNAME": bgcname,
                    "LCNAME": case_name
                })
        
        # Process other load analysis types
        analysis_types = {
            "MV": MV,
            "SM": SM,
            "THRSEV": THRSEV,
            "PO": PO,
            "THNS": THNS,
            "ULAT": ULAT
        }
        
        for analysis_type, bgcname in analysis_types.items():
            load_anal_entry = {
                "TYPE": analysis_type,
                "BGCNAME": bgcname if bgcname is not None else "UNCHANGED"
            }
            load_anal_list.append(load_anal_entry)
        
        return load_anal_list
    
    def _execute(self):
        """
        Automatically sends the BoundaryChangeAssignment to the system when created.
        """
        json_data = {"Assign": {}}
        
        boundary_data = {
            "bSPT": self.bSPT,
            "bSPR": self.bSPR,
            "bGSPR": self.bGSPR,
            "bCGLINK": self.bCGLINK,
            "bSSSF": self.bSSSF,
            "bPSSF": self.bPSSF,
            "bRLS": self.bRLS,
            "bWSSF": self.bWSSF,
            "bESSF": self.bESSF,
            "bCDOF": self.bCDOF,
            "vBOUNDARY": self.vBOUNDARY,
            "vLOADANAL": self.vLOADANAL
        }
        
        json_data["Assign"][str(self.ID)] = boundary_data
        
        # Execute the API call
        MidasAPI("PUT", "/db/bcct", json_data)
    
    def __str__(self):
        """
        String representation of the BoundaryChangeAssignment.
        """
        return f"BoundaryChangeAssignment(ID={self.ID}, Boundaries={len(self.vBOUNDARY)}, LoadAnalyses={len(self.vLOADANAL)})"
    
    def __repr__(self):
        """
        Detailed representation of the BoundaryChangeAssignment.
        """
        return (f"BoundaryChangeAssignment(ID={self.ID}, "
                f"bSPT={self.bSPT}, bCDOF={self.bCDOF}, "
                f"Boundaries={self.vBOUNDARY}, "
                f"LoadAnalyses={self.vLOADANAL})")



# Example usage:
"""
# Example 1: Basic boundary change assignment
boundary_assignment = BoundaryChangeAssignment(
    bSPT=True,
    bCDOF=True,
    vBOUNDARY=[["L1", "BG2"], ["L2", "BG1"]],
    ST_load_assignments=[["Self-weight", "L1"]],
    MV="L1"
)

# Example 2: Complex assignment with multiple settings
complex_assignment = BoundaryChangeAssignment(
    bSPT=True,
    bSPR=False,
    bGSPR=False,
    bCGLINK=False,
    bSSSF=False,
    bPSSF=False,
    bRLS=False,
    bCDOF=True,
    vBOUNDARY=[["L1", "BG2"], ["L2", "BG1"]],
    ST_load_assignments=[["Self-weight", "L1"], ["SIDL", "UNCHANGED"]],
    MV="UNCHANGED",
    SM="L2",
    THRSEV="L1",
    PO="UNCHANGED",
    THNS="UNCHANGED",
    ULAT="UNCHANGED"
)

# Example 3: Minimal assignment with defaults
minimal_assignment = BoundaryChangeAssignment(
    bSPT=True,
    vBOUNDARY=[["L1", "BG2"]]
)
"""


from ._mapi import MidasAPI

class Settlement:
    
    @classmethod
    def create(cls):
        """Creates Settlement Load in MIDAS Civil NX"""
        if cls.Group.data != []: cls.Group.create()
        if cls.Case.data != []: cls.Case.create()
    
    @classmethod
    def delete(cls):
        """Deletes Settlement load from MIDAS Civil NX and Python"""
        cls.Group.delete()
        cls.Case.delete()
    
    @classmethod
    def sync(cls):
        """Sync Settlement load from MIDAS Civil NX to Python"""
        cls.Group.sync()
        cls.Case.sync()
    
    class Group:
        """ 
        Parameters:
            name: Settlement group name (string)
            displacement: Settlement displacement value (number)
            node_list: List of node IDs to include in the group (array of integers)
            id: Group ID (optional, auto-generated if not provided)
        
        Examples:
            ```python
            Settlement.Group("SG1", 0.025, [100, 101])
            Settlement.Group("SG2", 0.015, [102, 103])
            ```
        """
        data = []

        def __init__(self, name, displacement, node_list, id=None):
            if id == None: id =""
            self.NAME = name
            self.SETTLE = displacement
            self.ITEMS = node_list
            if id == "": id = len(Settlement.Group.data) + 1
            self.ID = id
            
            Settlement.Group.data.append(self)

        @classmethod
        def json(cls):
            json = {"Assign": {}}
            for i in cls.data:
                json["Assign"][str(i.ID)] = {
                    "NAME": i.NAME,
                    "SETTLE": i.SETTLE,
                    "ITEMS": i.ITEMS
                }
            return json
        
        @staticmethod
        def create():
            MidasAPI("PUT", "/db/smpt", Settlement.Group.json())
        
        @staticmethod
        def get():
            return MidasAPI("GET", "/db/smpt")
        
        @classmethod
        def delete(cls):
            cls.data = []
            return MidasAPI("DELETE", "/db/smpt")
        
        @classmethod
        def sync(cls):
            cls.data = []
            a = cls.get()
            if a != {'message': ''}:
                for i in a['SMPT'].keys():
                    Settlement.Group(
                        a['SMPT'][i]['NAME'], 
                        a['SMPT'][i]['SETTLE'], 
                        a['SMPT'][i]['ITEMS'],
                        int(i)
                    )

    
    class Case:
        """
        
        Parameters:
            name: Settlement load case name (string)
            settlement_groups: List of settlement group names to include (array of strings, default [])
            factor: Settlement scale factor (number, default 1.0)
            min_groups: Minimum number of settlement groups (integer, default 1)
            max_groups: Maximum number of settlement groups (integer, default 1)
            desc: Description of the settlement case (string, default "")
            id: Case ID (optional, auto-generated if not provided)
        
        Examples:
            ```python
            Settlement.Case("SMLC1", ["SG1"], 1.2, 1, 1, "Foundation Settlement Case")
            Settlement.Case("SMLC2", ["SG1", "SG2"], 1.0, 1, 2, "Combined Settlement")
            ```
        """
        data = []

        def __init__(self, name, settlement_groups=[],factor=1.0, min_groups=1, max_groups=1, desc="", id=None):
            if id == None: id =""
            self.NAME = name
            self.DESC = desc
            self.FACTOR = factor
            self.MIN = min_groups
            self.MAX = max_groups
            self.ST_GROUPS = settlement_groups
            if id == "": id = len(Settlement.Case.data) + 1
            self.ID = id
            
            Settlement.Case.data.append(self)

        @classmethod
        def json(cls):
            json = {"Assign": {}}
            for i in cls.data:
                json["Assign"][str(i.ID)] = {
                    "NAME": i.NAME,
                    "DESC": i.DESC,
                    "FACTOR": i.FACTOR,
                    "MIN": i.MIN,
                    "MAX": i.MAX,
                    "ST_GROUPS": i.ST_GROUPS
                }
            return json
        
        @staticmethod
        def create():
            MidasAPI("PUT", "/db/smlc", Settlement.Case.json())
        
        @staticmethod
        def get():
            return MidasAPI("GET", "/db/smlc")
        
        @classmethod
        def delete(cls):
            cls.data = []
            return MidasAPI("DELETE", "/db/smlc")
        
        @classmethod
        def sync(cls):
            cls.data = []
            a = cls.get()
            if a != {'message': ''}:
                for i in a['SMLC'].keys():
                    Settlement.Case(
                        a['SMLC'][i]['NAME'],
                        a['SMLC'][i]['DESC'], 
                        a['SMLC'][i]['FACTOR'],
                        a['SMLC'][i]['MIN'],
                        a['SMLC'][i]['MAX'],
                        a['SMLC'][i]['ST_GROUPS'],
                        int(i)
                    )



from ._offsetSS import Offset
from ._offsetSS import _common

import numpy as np
from math import hypot
class _SS_TAP_PSC_12CELL(_common):
    def __init__(self,Name='',Shape='1CEL',Joint=[0,0,0,0,0,0,0,0],
                    HO1_I=0,HO2_I=0,HO21_I=0,HO22_I=0,HO3_I=0,HO31_I=0,
                    BO1_I=0,BO11_I=0,BO12_I=0,BO2_I=0,BO21_I=0,BO3_I=0,
                    HI1_I=0,HI2_I=0,HI21_I=0,HI22_I=0,HI3_I=0,HI31_I=0,HI4_I=0,HI41_I=0,HI42_I=0,HI5_I=0,
                    BI1_I=0,BI11_I=0,BI12_I=0,BI21_I=0,BI3_I=0,BI31_I=0,BI32_I=0,BI4_I=0,

                    HO1_J=0,HO2_J=0,HO21_J=0,HO22_J=0,HO3_J=0,HO31_J=0,
                    BO1_J=0,BO11_J=0,BO12_J=0,BO2_J=0,BO21_J=0,BO3_J=0,
                    HI1_J=0,HI2_J=0,HI21_J=0,HI22_J=0,HI3_J=0,HI31_J=0,HI4_J=0,HI41_J=0,HI42_J=0,HI5_J=0,
                    BI1_J=0,BI11_J=0,BI12_J=0,BI21_J=0,BI3_J=0,BI31_J=0,BI32_J=0,BI4_J=0,

                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=0):
                
        self.ID = id
        self.NAME = Name
        self.SHAPE = Shape
        self.TYPE = 'TAPERED'

        # print("***********")
        # print(Name,Shape,Joint)

        self.JO1=bool(Joint[0])
        self.JO2=bool(Joint[1])
        self.JO3=bool(Joint[2])
        self.JI1=bool(Joint[3])
        self.JI2=bool(Joint[4])
        self.JI3=bool(Joint[5])
        self.JI4=bool(Joint[6])
        self.JI5=bool(Joint[7])

        self.OFFSET = Offset
        self.USESHEAR = bool(useShear)
        self.USE7DOF = bool(use7Dof)

        self.HO1_I = HO1_I
        self.HO2_I = HO2_I
        self.HO21_I = HO21_I
        self.HO22_I= HO22_I
        self.HO3_I = HO3_I
        self.HO31_I = HO31_I

        self.BO1_I = BO1_I
        self.BO11_I = BO11_I
        self.BO12_I = BO12_I
        self.BO2_I = BO2_I
        self.BO21_I = BO21_I
        self.BO3_I = BO3_I

        self.HI1_I = HI1_I
        self.HI2_I = HI2_I
        self.HI21_I = HI21_I
        self.HI22_I = HI22_I
        self.HI3_I = HI3_I
        self.HI31_I = HI31_I
        self.HI4_I = HI4_I
        self.HI41_I = HI41_I
        self.HI42_I = HI42_I
        self.HI5_I = HI5_I

        self.BI1_I = BI1_I
        self.BI11_I = BI11_I
        self.BI12_I = BI12_I
        self.BI21_I = BI21_I
        self.BI3_I = BI3_I
        self.BI31_I = BI31_I
        self.BI32_I = BI32_I
        self.BI4_I = BI4_I




        self.HO1_J = HO1_J
        self.HO2_J = HO2_J
        self.HO21_J = HO21_J
        self.HO22_J= HO22_J
        self.HO3_J = HO3_J
        self.HO31_J = HO31_J

        self.BO1_J = BO1_J
        self.BO11_J = BO11_J
        self.BO12_J = BO12_J
        self.BO2_J = BO2_J
        self.BO21_J = BO21_J
        self.BO3_J = BO3_J

        self.HI1_J = HI1_J
        self.HI2_J = HI2_J
        self.HI21_J = HI21_J
        self.HI22_J = HI22_J
        self.HI3_J = HI3_J
        self.HI31_J = HI31_J
        self.HI4_J = HI4_J
        self.HI41_J = HI41_J
        self.HI42_J = HI42_J
        self.HI5_J = HI5_J

        self.BI1_J = BI1_J
        self.BI11_J = BI11_J
        self.BI12_J = BI12_J
        self.BI21_J = BI21_J
        self.BI3_J = BI3_J
        self.BI31_J = BI31_J
        self.BI32_J = BI32_J
        self.BI4_J = BI4_J

    
    def __str__(self):
         return f'  >  ID = {self.ID}   |  PSC 1-2 CELL SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                "SECTTYPE": "TAPERED",
                "SECT_NAME": sect.NAME,
                "SECT_BEFORE": {
                    "SHAPE": sect.SHAPE,
                    "TYPE" : 11,
                    "SECT_I": {
                        "vSIZE_PSC_A": [sect.HO1_I,sect.HO2_I,sect.HO21_I,sect.HO22_I,sect.HO3_I,sect.HO31_I],
                        "vSIZE_PSC_B": [sect.BO1_I,sect.BO11_I,sect.BO12_I,sect.BO2_I,sect.BO21_I,sect.BO3_I,],
                        "vSIZE_PSC_C": [sect.HI1_I,sect.HI2_I,sect.HI21_I,sect.HI22_I,sect.HI3_I,sect.HI31_I,sect.HI4_I,sect.HI41_I,sect.HI42_I,sect.HI5_I],
                        "vSIZE_PSC_D": [sect.BI1_I,sect.BI11_I,sect.BI12_I,sect.BI21_I,sect.BI3_I,sect.BI31_I,sect.BI32_I,sect.BI4_I],
                        "S_WIDTH" : sect.HO1_I
                    },
                    "SECT_J": {
                        "vSIZE_PSC_A": [sect.HO1_J,sect.HO2_J,sect.HO21_J,sect.HO22_J,sect.HO3_J,sect.HO31_J],
                        "vSIZE_PSC_B": [sect.BO1_J,sect.BO11_J,sect.BO12_J,sect.BO2_J,sect.BO21_J,sect.BO3_J,],
                        "vSIZE_PSC_C": [sect.HI1_J,sect.HI2_J,sect.HI21_J,sect.HI22_J,sect.HI3_J,sect.HI31_J,sect.HI4_J,sect.HI41_J,sect.HI42_J,sect.HI5_J],
                        "vSIZE_PSC_D": [sect.BI1_J,sect.BI11_J,sect.BI12_J,sect.BI21_J,sect.BI3_J,sect.BI31_J,sect.BI32_J,sect.BI4_J],
                        "S_WIDTH" : sect.HO1_J
                    },
                    "Y_VAR": 1,
                    "Z_VAR": 1,
                    "WARPING_CHK_AUTO_I": True,
                    "WARPING_CHK_AUTO_J": True,
                    "SHEAR_CHK": False,
                    "WARPING_CHK_POS_I": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "WARPING_CHK_POS_J": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "USE_WEB_THICK_SHEAR": [[True, True,True],[True,True,True]],
                    "WEB_THICK_SHEAR": [[0,0,0],[0,0,0]],
                    "USE_WEB_THICK": [True,True],
                    "WEB_THICK": [0,0],
                    "USE_SYMMETRIC": False,
                    "USE_SMALL_HOLE": False,
                    "USE_USER_DEF_MESHSIZE": False,
                    "USE_USER_INTPUT_STIFF": False,
                    "PSC_OPT1": "",
                    "PSC_OPT2": "",
                    "JOINT": [sect.JO1,sect.JO2,sect.JO3,sect.JI1,sect.JI2,sect.JI3,sect.JI4,sect.JI5]
                }
            }
        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    

    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):
        #--- PSC 1,2 CELL -------------------
        vA_I = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_A']
        vB_I = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_B']
        vC_I = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_C']
        vD_I = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_D']

        vA_J = js['SECT_BEFORE']['SECT_J']['vSIZE_PSC_A']
        vB_J = js['SECT_BEFORE']['SECT_J']['vSIZE_PSC_B']
        vC_J = js['SECT_BEFORE']['SECT_J']['vSIZE_PSC_C']
        vD_J = js['SECT_BEFORE']['SECT_J']['vSIZE_PSC_D']

        joint = js['SECT_BEFORE']['JOINT']
        return _SS_TAP_PSC_12CELL(name,shape,joint,
                            *vA_I,*vB_I,*vC_I,*vD_I,
                            *vA_J,*vB_J,*vC_J,*vD_J,
                            offset,uShear,u7DOF,id)
    
    def _centerLine(shape,end,*args):
        if shape.SHAPE in ['1CEL','2CEL'] :
            if end:
                HO1,HO2,HO21,HO22,HO3,HO31 = shape.HO1_J,shape.HO2_J,shape.HO21_J,shape.HO22_J,shape.HO3_J,shape.HO31_J
                BO1,BO11,BO12,BO2,BO21,BO3 = shape.BO1_J,shape.BO11_J,shape.BO12_J,shape.BO2_J,shape.BO21_J,shape.BO3_J

                HI1,HI2,HI21,HI22,HI3,HI31,HI4,HI41,HI42,HI5 = shape.HI1_J,shape.HI2_J,shape.HI21_J,shape.HI22_J,shape.HI3_J,shape.HI31_J,shape.HI4_J,shape.HI41_J,shape.HI42_J,shape.HI5_J
                BI1,BI11,BI12,BI21,BI3,BI31,BI32,BI4 = shape.BI1_J,shape.BI11_J,shape.BI12_J,shape.BI21_J,shape.BI3_J,shape.BI31_J,shape.BI32_J,shape.BI4_J
                
            
            else:
                HO1,HO2,HO21,HO22,HO3,HO31 = shape.HO1_I,shape.HO2_I,shape.HO21_I,shape.HO22_I,shape.HO3_I,shape.HO31_I
                BO1,BO11,BO12,BO2,BO21,BO3 = shape.BO1_I,shape.BO11_I,shape.BO12_I,shape.BO2_I,shape.BO21_I,shape.BO3_I

                HI1,HI2,HI21,HI22,HI3,HI31,HI4,HI41,HI42,HI5 = shape.HI1_I,shape.HI2_I,shape.HI21_I,shape.HI22_I,shape.HI3_I,shape.HI31_I,shape.HI4_I,shape.HI41_I,shape.HI42_I,shape.HI5_I
                BI1,BI11,BI12,BI21,BI3,BI31,BI32,BI4 = shape.BI1_I,shape.BI11_I,shape.BI12_I,shape.BI21_I,shape.BI3_I,shape.BI31_I,shape.BI32_I,shape.BI4_I

                
                
            JO1,JO2,JO3,JI1,JI2,JI3,JI4,JI5 = shape.JO1,shape.JO2,shape.JO3,shape.JI1,shape.JI2,shape.JI3,shape.JI4,shape.JI5

            sect_cg_LT = [-(BO1+BO2+BO3),0]
            sect_cg_RB = [(BO1+BO2+BO3),-(HO1+HO2+HO3)]
            sect_cg_CC = [0,-(HO1+HO2+HO3)/2]




            def perpendicular_point(pt1,pt2,point, l=0):
                """Function to get orthogonal point on line (x1,y1)-(x2,y2) from point (x3,y3). Enter l=0 for point 3 in between 1 & 2.  Enter l=1 for other scenario."""
                x1,y1 = pt1
                x2,y2 = pt2
                x3,y3 = point

                if x2 != x1:
                    m = (y2 - y1) / (x2 - x1)
                    c = y1 - m * x1
                    x_perp = (x3 + m * (y3 - c)) / (1 + m**2)
                    y_perp = m * x_perp + c
                else:
                    x_perp, y_perp = x1, y3
                
                thk = ((x3 - x_perp)**2 + (y3 - y_perp)**2)**0.5
                return (x_perp, y_perp),thk
            
            def distance_point(pt1,pt2):
                x1,y1 = pt1
                x2,y2 = pt2

                return hypot((x1-x2),(y1-y2))


            HTI = HI1+HI2+HI3+HI4+HI5
            HTO = HO1+HO2+HO3
            slope = (HTI-HTO)/(BO1+BO2+BO3)


            pt1 = (0,0)
            pt2 = (0,-HI1)
            pt5 = (-BI1,-HI1-HI2)
            pt7 = (-BI3,-HI1-HI2-HI3)
            pt10 = (0,-HI1-HI2-HI3-HI4)
            pt11 = (0,-HTI)

            pt12 = (-BO3,-HTI)
            pt14 = (-BO3-BO2,(-HTI+HTO)-HO1-HO2)
            pt17 = (-BO3-BO2-BO1,(-HTI+HTO)-HO1)
            pt18 = (-BO3-BO2-BO1,(-HTI+HTO))

            pt3 = (-BI11,-HI1-HI21)
            pt4 = (-BI12,-HI1-HI22)

            pt6 = (-BI21,-HI1-HI2-HI31)

            pt8 = (-BI32,-HI1-HI2-HI3-HI4+HI42)
            pt9 = (-BI31,-HI1-HI2-HI3-HI4+HI41)

            pt13 = (-BO3-BO21,-HO1-HO2-HO3+HO31)

            pt15 = (-BO3-BO2-BO1+BO12,(-HTI+HTO)-HO1-HO22)
            pt16 = (-BO3-BO2-BO1+BO11,(-HTI+HTO)-HO1-HO21)


            pt016 = (-BO3-BO2-BO1+BO11,slope*(-BO3-BO2-BO1+BO11))
            pt015 = (-BO3-BO2-BO1+BO12,slope*(-BO3-BO2-BO1+BO12))
            pt014 = (-BO3-BO2,slope*(-BO3-BO2))
            pt05 = (-BI1,slope*(-BI1))
            pt04 = (-BI12,slope*(-BI12))
            pt03 = (-BI11,slope*(-BI11))

            #----------------------------- THICKNESS COMPUTATION --------------

            THI1 = HI1
            TJI1 = distance_point(pt3,pt03)
            TJI2 = distance_point(pt4,pt04)
            THI2 = distance_point(pt5,pt05)
            THO2 = distance_point(pt14,pt014)
            THO1 = distance_point(pt17,pt18)
            THI3 = (HI4+HI5)
            THI5 = HI5
            TJO1 = distance_point(pt16,pt016)
            TJO2 = distance_point(pt15,pt015)
            TJI4 = (HI5+HI42)
            TJI5 = (HI5+HI41)

            # -------------------   CENTER LINE POINTS ------------------

            cp1 = (0,-HI1/2)


            # cp2 = (-BI11,(-HI1-HI21)/2) #JI1
            # cp2 = np.add(pt3,pt03)/2
            cp2 = (pt03[0],pt03[1]-0.5*THI1)

            # cp3 = (-BI12,(-HI1-HI22)/2) #JI2
            # cp3 = np.add(pt4,pt04)/2
            cp3 = (pt04[0],pt04[1]-0.5*THI1)
            # cp4 = (-BI1,(-HI1-HI2)/2)
            # cp4 = np.add(pt5,pt05)/2
            cp4 = (pt05[0],pt05[1]-0.5*THI1)

            # cp5 = (-BO3-BO2-BO1,-HO1/2)
            cp5 = np.add(pt17,pt18)/2
            # cp6 = (-BO3-BO2-BO1+BO11,(-HO1-HO21)/2) #JO1
            # cp6 = np.add(pt16,pt016)/2
            cp6 = (pt016[0],pt016[1]-0.5*THO1)
            # cp7 = (-BO3-BO2-BO1+BO12,(-HO1-HO22)/2) #JO2
            # cp7 = np.add(pt15,pt015)/2
            cp7 = (pt015[0],pt015[1]-0.5*THO1)
            # cp8 = (-BO3-BO2,(-HO1-HO2)/2)
            # cp8 = np.add(pt14,pt014)/2
            cp8 = (pt014[0],pt014[1]-0.5*THO1)

            cp9 = np.add(cp8,cp4)/2

            cp10 = np.add(pt5,pt14)/2

            cp17 = np.add(pt10,pt11)/2
            cp16 = (-BI31,(-HI1-HI2-HI3-HI4+HI41-HTI)/2) #JI5
            cp15 = (-BI32,(-HI1-HI2-HI3-HI4+HI42-HTI)/2) #JI4
            cp14 = (-BI3,(-HI1-HI2-HI3-HTI)/2)
            cp13 = np.add(pt7,pt12)/2


            tpt,TJI3 = perpendicular_point(pt13,pt14,pt6)
            cp11 = np.add(tpt,pt6)/2  #JI3


            tpt,TJO3 = perpendicular_point(pt7,pt6,pt13)
            cp12 = np.add(pt7,pt13)/2 #JO3


            # if cp12[1] < cp13[1] : 
            #     print("JO3 invalid")


            


            TX_INT_FLANGE = (THO2+THI2)/2
            tpt,THIO2 = perpendicular_point(pt13,pt14,pt5)
            TY_INT_FLANGE = THIO2

            TX_BOT_JUNC = hypot(*np.subtract(pt7,pt12))

            tpt,TY_BOT_JUNC = perpendicular_point(pt13,pt14,pt7)
            # TY_BOT_JUNC = (TY_BOT_JUNC+hypot(*np.subtract(pt7,pt12)))/2

            TJO3 = max(TJO3,TY_BOT_JUNC)


            top_flange_point =[]
            top_flange_line = []
            top_flange_thk = []
            top_flange_thk_off = []
            q=1
            top_flange_point.append(cp1)
            if JI1 : 
                top_flange_point.append(cp2)
                top_flange_line.append([q,q+1])
                top_flange_thk.append([THI1,TJI1])
                top_flange_thk_off.append([0,(TJI1-THI1)*0.5])
                q+=1
            if JI2 : 
                top_flange_point.append(cp3)
                top_flange_line.append([q,q+1])
                q+=1
                if JI1:
                    top_flange_thk.append([TJI1,TJI2])
                    top_flange_thk_off.append([(TJI1-THI1)*0.5,(TJI2-THI1)*0.5])
                else :
                    top_flange_thk.append([THI1,TJI2])
                    top_flange_thk_off.append([0,(TJI2-THI1)*0.5])
            top_flange_point.append(cp4)
            top_flange_line.append([q,q+1])
            q+=1

            if JI2:
                top_flange_thk.append([TJI2,THI2])
                top_flange_thk_off.append([(TJI2-THI1)*0.5,(THI2-THI1)*0.5])
            elif JI1:
                top_flange_thk.append([TJI1,THI2])
                top_flange_thk_off.append([(TJI1-THI1)*0.5,(THI2-THI1)*0.5])
            else :
                top_flange_thk.append([THI1,THI2])
                top_flange_thk_off.append([0,(THI2-THI1)*0.5])


            top_flange_line.append([q,q+1]) # TO CONNECT WITH WEB
            q+=1
            top_flange_thk.append([THI2,TX_INT_FLANGE]) 
            top_flange_thk_off.append([(THI2-THI1)*0.5,(TX_INT_FLANGE-THI1)*0.5])


            # print(top_flange_thk)

            web_point =[]
            web_line =[]
            web_thk = []
            web_thk_off = []

            web_point.append(cp9)
            web_point.append(cp10)
            web_line.append([q,q+1])
            q+=1
            web_thk.append([TY_INT_FLANGE,THIO2])
            web_thk_off.append([0,0])
            if JI3 : 
                web_point.append(cp11)
                web_line.append([q,q+1])
                q+=1
                web_thk.append([THIO2,TJI3])
                web_thk_off.append([0,0])
            if JO3 : #JO3 and cp12[0] < cp13[0]: 
                web_point.append(cp12)
                web_line.append([q,q+1])
                if JI3 : 
                    web_thk.append([TJI3,TJO3])
                    web_thk_off.append([0,0])
                else:
                    web_thk.append([THIO2,TJO3])
                    web_thk_off.append([0,0])
                q+=1

            web_point.append(cp13)
            web_line.append([q,q+1])
            if JO3:
                web_thk.append([TJO3,TY_BOT_JUNC])
                web_thk_off.append([0,0])
            elif JI3:
                web_thk.append([TJI3,TY_BOT_JUNC])
                web_thk_off.append([0,0])
            else:
                web_thk.append([THIO2,TY_BOT_JUNC])
                web_thk_off.append([0,0])
            q+=1


            bottom_flange_point =[]
            bottom_flange_line =[]
            bottom_flange_thk = []
            bottom_flange_thk_off = []

            bottom_flange_line.append([q,q+1]) # TO CONNECT WITH WEB
            bottom_flange_thk.append([TY_BOT_JUNC,THI3])
            bottom_flange_thk_off.append([0,0])
            q+=1
            bottom_flange_point.append(cp14)
            if JI4 : 
                bottom_flange_point.append(cp15)
                bottom_flange_line.append([q,q+1])
                q+=1
                bottom_flange_thk.append([THI3,TJI4])
                bottom_flange_thk_off.append([0,0])
            if JI5 : 
                bottom_flange_point.append(cp16)
                bottom_flange_line.append([q,q+1])
                q+=1
                if JI4:
                    bottom_flange_thk.append([TJI4,TJI5])
                    bottom_flange_thk_off.append([0,0])
                else:
                    bottom_flange_thk.append([THI3,TJI5])
                    bottom_flange_thk_off.append([0,0])
            bottom_flange_point.append(cp17)
            bottom_flange_line.append([q,q+1])
            n_cp17 = q+1
            if JI5:
                bottom_flange_thk.append([TJI5,THI5])
                bottom_flange_thk_off.append([0,0])
            elif JI4:
                bottom_flange_thk.append([TJI4,THI5])
                bottom_flange_thk_off.append([0,0])
            else:
                bottom_flange_thk.append([THI3,THI5])
                bottom_flange_thk_off.append([0,0])
            q+=1 


            top_cantelever_point =[]
            top_cantelever_line =[]
            top_cantelever_thk =[]
            top_cantelever_thk_off =[]

            top_cantelever_line.append([len(top_flange_point)+1,q+1]) # TO CONNECT WITH WEB
            top_cantelever_thk.append([TX_INT_FLANGE,THO2])
            top_cantelever_thk_off.append([(TX_INT_FLANGE-THO1)*0.5,(THO2-THO1)*0.5])
            q+=1
            top_cantelever_point.append(cp8)
            if JO2 : 
                top_cantelever_point.append(cp7)
                top_cantelever_line.append([q,q+1])
                q+=1
                top_cantelever_thk.append([THO2,TJO2])
                top_cantelever_thk_off.append([(THO2-THO1)*0.5,(TJO2-THO1)*0.5])
            if JO1 : 
                top_cantelever_point.append(cp6)
                top_cantelever_line.append([q,q+1])
                q+=1
                if JO2:
                    top_cantelever_thk.append([TJO2,TJO1])
                    top_cantelever_thk_off.append([(TJO2-THO1)*0.5,(TJO1-THO1)*0.5])
                else:
                    top_cantelever_thk.append([THO2,TJO1])
                    top_cantelever_thk_off.append([(THO2-THO1)*0.5,(TJO1-THO1)*0.5])
            top_cantelever_point.append(cp5)
            top_cantelever_line.append([q,q+1])
            if JO1:
                top_cantelever_thk.append([TJO1,THO1])
                top_cantelever_thk_off.append([(TJO1-THO1)*0.5,0])
            elif JO2:
                top_cantelever_thk.append([TJO2,THO1])
                top_cantelever_thk_off.append([(TJO2-THO1)*0.5,0])
            else:
                top_cantelever_thk.append([THO2,THO1])
                top_cantelever_thk_off.append([(THO2-THO1)*0.5,0])


            final_points = top_flange_point+web_point+bottom_flange_point+top_cantelever_point
            final_line = top_flange_line+web_line+bottom_flange_line+top_cantelever_line
            final_thk = top_flange_thk+web_thk+bottom_flange_thk+top_cantelever_thk
            final_thk_off = top_flange_thk_off+web_thk_off+bottom_flange_thk_off+top_cantelever_thk_off

            n_points = len(final_points)

            # print(final_points)

            sect_shape = final_points   # SYMMETRY
            for i in range(n_points):
                sect_shape.append((-final_points[i][0],final_points[i][1]))


            sect_lin_con = final_line
            sect_thk = final_thk
            sect_thk_off = final_thk_off
            

            for q in range(len(final_line)):    # SYMMETRY
                sect_thk.append([final_thk[q][1],final_thk[q][0]])
                sect_thk_off.append([final_thk_off[q][1],final_thk_off[q][0]])
                sect_lin_con.append([final_line[q][1]+n_points,final_line[q][0]+n_points])


            if shape.SHAPE == '2CEL':
                sect_lin_con.append([1,n_cp17])
                sect_thk.append([2*BI4,2*BI4])
                sect_thk_off.append([0,0])
                
            # sect_thk_off = [0 for _ in sect_thk]
            
        

        sect_cg = (sect_cg_LT,sect_cg_CC,sect_cg_RB)
        return sect_shape, sect_thk ,sect_thk_off, sect_cg , sect_lin_con
    










from ._offsetSS import Offset
from ._offsetSS import _common
import math

class _SS_TAPERED_DBUSER(_common):

    """ Create Standard USER DEFINED sections"""

    def __init__(self,Name='',Shape='',params_I:list=[],params_J:list=[],Offset=Offset(),useShear=True,use7Dof=False,id:int=0):  
        """ Shape = 'SB' 'SR' for rectangle \n For cylinder"""
        self.ID = id
        self.NAME = Name
        self.TYPE = 'TAPERED'
        self.SHAPE = Shape
        self.PARAMS_I = params_I
        self.PARAMS_J = params_J
        self.OFFSET = Offset
        self.USESHEAR = useShear
        self.USE7DOF = use7Dof
        self.DATATYPE = 2
    
    def __str__(self):
         return f'  >  ID = {self.ID}   |  USER DEFINED STANDARD SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                "SECTTYPE": sect.TYPE,
                "SECT_NAME": sect.NAME,
                "SECT_BEFORE": {
                    "SHAPE": sect.SHAPE,
                    "TYPE": sect.DATATYPE,
                    "SECT_I": {
                        "vSIZE": sect.PARAMS_I
                    },
                    "SECT_J": {
                        "vSIZE": sect.PARAMS_J
                    }
                }
            }
        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    
    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):
        return _SS_TAPERED_DBUSER(name,shape,js['SECT_BEFORE']['SECT_I']['vSIZE'],js['SECT_BEFORE']['SECT_J']['vSIZE'],offset,uShear,u7DOF,id)
    
    def _centerLine(shape,end,*args):
        if end: 
            shape.PARAMS = shape.PARAMS_J
            # print(' J end taken')
        else: 
            # print(' I end taken')
            shape.PARAMS = shape.PARAMS_I

        if shape.SHAPE == 'SB' :
            H,B = shape.PARAMS[:2]

            sect_lin_con = [[1,2],[3,1]]

            sect_cg_LT = [-B/2,H/2]
            sect_cg_CC = [0,0]
            sect_cg_RB = [B/2,-H/2]

            if H > B :
                sect_shape = [[0,0],[0,H/2],[0,-H/2],[0,H/4],[0,-H/4]]
                sect_thk = [B,B,B,B]
                sect_thk_off = [0,0,0,0]
            else : 
                sect_shape = [[0,0],[B/2,0],[-B/2,0],[B/4,0],[-B/4,0]]
                sect_thk = [H,H,H,H]
                sect_thk_off = [0,0,0,0]

        elif shape.SHAPE == 'L' :
            H,B,tw,tf = shape.PARAMS[:4]

            sect_cg_LT = [0,0]
            sect_cg_CC = [(H*tw*tw+B*B*tf)/(2*(B*tw+H*tf)),-(H*H*tw+B*tf*tf)/(2*(B*tw+H*tf))]
            sect_cg_RB = [B,-H]

            # sect_shape = [[0.5*tw,-H],[0.5*tw,-0.5*tf],[B,-0.5*tf]]
            sect_shape = [[0,-H],[0,0],[B,0]]
            sect_lin_con = [[3,2],[2,1]]
            sect_thk = [tw,tf]
            # sect_thk_off = [0,0]
            sect_thk_off = [tw/2,tf/2]
        
        elif shape.SHAPE == 'C' :
            H,B1,tw,tf1,B2,tf2 = shape.PARAMS[:6]
            if B2 == 0 : B2 = B1
            if tf2 == 0 : tf2 = tf1

            sect_cg_LT = [0,0]
            sect_cg_CC = [(B1+B2)*0.2,-H*0.5]
            sect_cg_RB = [max(B1,B2),-H]

            # sect_shape = [[0.5*tw,-0.5*tf1],[B1,-0.5*tf1],[0.5*tw,-H+0.5*tf2],[B2,-H+0.5*tf2]]
            sect_shape = [[0,0],[B1,0],[0,-H],[B2,-H]]
            sect_lin_con = [[2,1],[1,3],[3,4]]
            sect_thk = [tf1,tw,tf2]
            # sect_thk_off = [0,0,0]
            sect_thk_off = [tf1/2,tw/2,tf2/2]

        elif shape.SHAPE == 'H' :
            H,B1,tw,tf1,B2,tf2,r1,r2 = shape.PARAMS[:8]
            if B2 == 0 : B2 = B1
            if tf2 == 0 : tf2 = tf1

            sect_cg_LT = [-0.5*max(B1,B2),0.5*H]
            sect_cg_CC = [0,0]
            sect_cg_RB = [0.5*max(B1,B2),-0.5*H]

            sect_shape = [[-0.5*B1,0.5*(H-tf1)],[0,0.5*(H-tf1)],[0.5*B1,0.5*(H-tf1)],[-0.5*B2,-0.5*(H-tf2)],[0,-0.5*(H-tf2)],[0.5*B2,-0.5*(H-tf2)]]
            sect_lin_con = [[2,1],[3,2],[2,5],[4,5],[5,6]]
            sect_thk = [tf1,tf1,tw,tf2,tf2]
            sect_thk_off = [0,0,0,0,0]
        
        elif shape.SHAPE == 'T' :
            H,B,tw,tf = shape.PARAMS[:4]

            sect_cg_LT = [-B*0.5,0]
            sect_cg_CC = [0,-H*0.3]
            sect_cg_RB = [B*0.5,-H]

            sect_shape = [[-0.5*B,-0.5*tf],[0,-0.5*tf],[0.5*B,-0.5*tf],[0,-H]]
            sect_lin_con = [[2,1],[3,2],[2,4]]
            sect_thk = [tf,tf,tw]
            sect_thk_off = [0,0,0]

        elif shape.SHAPE == 'B' :
            H,B,tw,tf1,C,tf2 = shape.PARAMS[:6]
            if tf2 == 0 : tf2 = tf1

            sect_cg_LT = [-0.5*B,0.5*H]
            sect_cg_CC = [0,0]
            sect_cg_RB = [0.5*B,-0.5*H]

            # sect_shape = [[0.5*(B-tw),0.5*(H-tf1)],[-0.5*(B-tw),0.5*(H-tf1)],[-0.5*(B-tw),-0.5*(H-tf2)],[0.5*(B-tw),-0.5*(H-tf2)]]
            sect_shape = [[0.5*B,0.5*H],[-0.5*B,0.5*H],[-0.5*B,-0.5*H],[0.5*B,-0.5*H]]

            sect_lin_con = [[1,2],[2,3],[3,4],[4,1]]
            sect_thk = [tf1,tw,tf2,tw]
            # sect_thk_off = [0,0,0,0]
            sect_thk_off = [0.5*tf1,0.5*tw,0.5*tf2,0.5*tw]
        
        elif shape.SHAPE == 'P' :
            D,tw = shape.PARAMS[:2]

            # R = 0.5*(D-tw)
            R = 0.5*D

            sect_cg_LT = [-R,R]
            sect_cg_CC = [0,0]
            sect_cg_RB = [R,-R]

            sect_shape = []
            sect_lin_con = []
            sect_thk = []
            sect_thk_off = []

            n = 16
            for i in range(n):
                sect_shape.append([R*math.sin(i*2*math.pi/n),R*math.cos(i*2*math.pi/n)])
                sect_lin_con.append([i+1,i+2])
                sect_thk.append(tw)
                sect_thk_off.append(-0.5*tw)
            sect_lin_con[-1] = [i+1,1]



        sect_cg = (sect_cg_LT,sect_cg_CC,sect_cg_RB)

        return sect_shape, sect_thk ,sect_thk_off, sect_cg , sect_lin_con